import bottleneck as bn
import dask.array as darray
import numpy as np
import properscoring
import xarray as xr

from .contingency import Contingency, _get_category_bounds
from .utils import (
    _add_as_coord,
    _bool_to_int,
    _check_identical_xr_types,
    _fail_if_dim_empty,
    _get_bin_centers,
    _keep_nans_masked,
    _preprocess_dims,
    _stack_input_if_needed,
    histogram,
    suppress_warnings,
)

__all__ = [
    "brier_score",
    "crps_ensemble",
    "crps_gaussian",
    "crps_quadrature",
    "threshold_brier_score",
    "rank_histogram",
    "discrimination",
    "reliability",
    "rps",
    "roc",
]

FORECAST_PROBABILITY_DIM = "forecast_probability"


def crps_gaussian(observations, mu, sig, dim=None, weights=None, keep_attrs=False):
    """Continuous Ranked Probability Score with a Gaussian distribution.

    Parameters
    ----------
    observations : xarray.Dataset or xarray.DataArray
        The observations or set of observations.
    mu : xarray.Dataset or xarray.DataArray
        The mean of the forecast normal distribution.
    sig : xarray.Dataset or xarray.DataArray
        The standard deviation of the forecast distribution.
    dim : str or list of str, optional
        Dimension over which to compute mean after computing ``crps_gaussian``.
        Defaults to None implying averaging over all dimensions.
    weights : xr.DataArray with dimensions from dim, optional
        Weights for `weighted.mean(dim)`.
        Defaults to None, such that no weighting is applied.
    keep_attrs : bool
        If True, the attributes (attrs) will be copied
        from the first input to the new one.
        If False (default), the new object will
        be returned without attributes.

    Returns
    -------
    xarray.Dataset or xarray.DataArray reduced by dimension dim

    Examples
    --------
    >>> observations = xr.DataArray(np.random.normal(size=(3,3)),
    ...                             coords=[('x', np.arange(3)),
    ...                                     ('y', np.arange(3))])
    >>> forecasts = xr.DataArray(np.random.normal(size=(3,3,3)),
    ...                          coords=[('x', np.arange(3)),
    ...                                  ('y', np.arange(3)),
    ...                                  ('member', np.arange(3))])
    >>> mu = forecasts.mean('member')
    >>> sig = forecasts.std('member')
    >>> crps_gaussian(observations, mu, sig, dim='x')
    <xarray.DataArray (y: 3)>
    array([1.0349773 , 0.36521376, 0.39017126])
    Coordinates:
      * y        (y) int64 0 1 2

    See Also
    --------
    properscoring.crps_gaussian
    """
    # check if same dimensions
    if isinstance(mu, (int, float)):
        mu = xr.DataArray(mu)
    if isinstance(sig, (int, float)):
        sig = xr.DataArray(sig)
    if mu.dims != observations.dims:
        observations, mu = xr.broadcast(observations, mu)
    if sig.dims != observations.dims:
        observations, sig = xr.broadcast(observations, sig)
    res = xr.apply_ufunc(
        properscoring.crps_gaussian,
        observations,
        mu,
        sig,
        input_core_dims=[[], [], []],
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=keep_attrs,
    )
    if weights is not None:
        return res.weighted(weights).mean(dim, keep_attrs=keep_attrs)
    else:
        return res.mean(dim, keep_attrs=keep_attrs)


def crps_quadrature(
    observations,
    cdf_or_dist,
    xmin=None,
    xmax=None,
    tol=1e-6,
    dim=None,
    weights=None,
    keep_attrs=False,
):
    """Continuous Ranked Probability Score with numerical integration
    of the normal distribution.

    Parameters
    ----------
    observations : xarray.Dataset or xarray.DataArray
        Observations associated with the forecast distribution ``cdf_or_dist``.
    cdf_or_dist : callable or scipy.stats.distribution
        Function which returns the cumulative density of the forecast
        distribution at value x.
    xmin, xmax, tol: see properscoring.crps_quadrature
    dim : str or list of str, optional
        Dimension over which to compute mean after computing ``crps_quadrature``.
        Defaults to None implying averaging over all dimensions.
    weights : xr.DataArray with dimensions from dim, optional
        Weights for `weighted.mean(dim)`.
        Defaults to None, such that no weighting is applied.
    keep_attrs : bool
        If True, the attributes (attrs) will be copied
        from the first input to the new one.
        If False (default), the new object will
        be returned without attributes.

    Returns
    -------
    xarray.Dataset or xarray.DataArray

    Examples
    --------
    >>> observations = xr.DataArray(np.random.normal(size=(3, 3)),
    ...                             coords=[('x', np.arange(3)),
    ...                                     ('y', np.arange(3))])
    >>> from scipy.stats import norm
    >>> xs.crps_quadrature(observations, norm, dim="x")
    <xarray.DataArray (y: 3)>
    array([0.80280921, 0.31818197, 0.32364912])
    Coordinates:
      * y        (y) int64 0 1 2

    See Also
    --------
    properscoring.crps_quadrature
    """
    res = xr.apply_ufunc(
        properscoring.crps_quadrature,
        observations,
        cdf_or_dist,
        xmin,
        xmax,
        tol,
        input_core_dims=[[], [], [], [], []],
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=keep_attrs,
    )
    if weights is not None:
        return res.weighted(weights).mean(dim, keep_attrs=keep_attrs)
    else:
        return res.mean(dim, keep_attrs=keep_attrs)


def crps_ensemble(
    observations,
    forecasts,
    member_weights=None,
    issorted=False,
    member_dim="member",
    dim=None,
    weights=None,
    keep_attrs=False,
):
    """Continuous Ranked Probability Score with the ensemble distribution.

    Parameters
    ----------
    observations : xarray.Dataset or xarray.DataArray
        The observations or set of observations.
    forecasts : xarray.Dataset or xarray.DataArray
        Forecast with required member dimension ``member_dim``.
    member_weights : xarray.Dataset or xarray.DataArray
        If provided, the CRPS is calculated exactly with the assigned
        probability weights to each forecast. Weights should be positive,
        but do not need to be normalized. By default, each forecast is
        weighted equally.
    issorted : bool, optional
        Optimization flag to indicate that the elements of `ensemble` are
        already sorted along `axis`.
    member_dim : str, optional
        Name of ensemble member dimension. By default, 'member'.
    dim : str or list of str, optional
        Dimension over which to compute mean after computing ``crps_ensemble``.
        Defaults to None implying averaging over all dimensions.
    weights : xr.DataArray with dimensions from dim, optional
        Weights for `weighted.mean(dim)`.
        Defaults to None, such that no weighting is applied.
    keep_attrs : bool
        If True, the attributes (attrs) will be copied
        from the first input to the new one.
        If False (default), the new object will
        be returned without attributes.

    Returns
    -------
    xarray.Dataset or xarray.DataArray

    Examples
    --------
    >>> observations = xr.DataArray(np.random.normal(size=(3,3)),
    ...                             coords=[('x', np.arange(3)),
    ...                                     ('y', np.arange(3))])
    >>> forecasts = xr.DataArray(np.random.normal(size=(3,3,3)),
    ...                          coords=[('x', np.arange(3)),
    ...                                  ('y', np.arange(3)),
    ...                                  ('member', np.arange(3))])
    >>> crps_ensemble(observations, forecasts, dim='x')
    <xarray.DataArray (y: 3)>
    array([1.04497153, 0.48997746, 0.47994095])
    Coordinates:
      * y        (y) int64 0 1 2

    See Also
    --------
    properscoring.crps_ensemble
    """
    res = xr.apply_ufunc(
        properscoring.crps_ensemble,
        observations,
        forecasts,
        input_core_dims=[[], [member_dim]],
        kwargs={"axis": -1, "issorted": issorted, "weights": member_weights},
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=keep_attrs,
    )
    if weights is not None:
        return res.weighted(weights).mean(dim, keep_attrs=keep_attrs)
    else:
        return res.mean(dim, keep_attrs=keep_attrs)


def brier_score(
    observations,
    forecasts,
    member_dim="member",
    fair=False,
    dim=None,
    weights=None,
    keep_attrs=False,
):
    """Calculate Brier score (BS).

    .. math:
        BS(p, k) = (p_1 - k)^{2}

    Parameters
    ----------
    observations : xarray.Dataset or xarray.DataArray
        The observations or set of observations of the event.
        Data should be boolean or logical \
        (True or 1 for event occurance, False or 0 for non-occurance).
    forecasts : xarray.Dataset or xarray.DataArray
        The forecast likelihoods of the event.
        If ``fair==False``, forecasts should be between 0 and 1 without a dimension
        ``member_dim`` or should be boolean (True,False) or binary (0, 1) containing a
        member dimension (probabilities will be internally calculated by
        ``.mean(member_dim))``. If ``fair==True``, forecasts must be boolean
        (True,False) or binary (0, 1) containing dimension ``member_dim``.
    member_dim : str, optional
        Name of ensemble member dimension. By default, 'member'.
    fair: boolean
        Apply ensemble member-size adjustment for unbiased, fair metric;
        see Ferro (2013). Defaults to False.
    dim : str or list of str, optional
        Dimension over which to compute mean after computing ``brier_score``.
        Defaults to None implying averaging over all dimensions.
    weights : xr.DataArray with dimensions from dim, optional
        Weights for `weighted.mean(dim)`.
        Defaults to None, such that no weighting is applied.
    keep_attrs : bool
        If True, the attributes (attrs) will be copied
        from the first input to the new one.
        If False (default), the new object will
        be returned without attributes.

    Returns
    -------
    xarray.Dataset or xarray.DataArray

    Examples
    --------
    >>> observations = xr.DataArray(np.random.normal(size=(3, 3)),
    ...                             coords=[('x', np.arange(3)),
    ...                                     ('y', np.arange(3))])
    >>> forecasts = xr.DataArray(np.random.normal(size=(3, 3, 3)),
    ...                          coords=[('x', np.arange(3)),
    ...                                  ('y', np.arange(3)),
    ...                                  ('member', np.arange(3))])
    >>> xs.brier_score(observations > .5,
    ...                (forecasts > .5).mean('member'),
    ...                dim="y")
    <xarray.DataArray (x: 3)>
    array([0.37037037, 0.14814815, 0.51851852])
    Coordinates:
      * x        (x) int64 0 1 2

    See Also
    --------
    properscoring.brier_score

    References
    ----------
    * Gneiting, Tilmann, and Adrian E Raftery. “Strictly Proper Scoring Rules,
      Prediction, and Estimation.” Journal of the American Statistical
      Association 102, no. 477 (March 1, 2007): 359–78.
      https://doi.org/10/c6758w.
    * Brier, Glenn W. "VERIFICATION OF FORECASTS EXPRESSED IN TERMS OF PROBABILITY."
      Monthly Weather Review, 78(1): 1-3
      https://journals.ametsoc.org/doi/abs/10.1175/1520-0493%281950%29078%3C0001%3AVOFEIT%3E2.0.CO%3B2
    * C. A. T. Ferro. Fair scores for ensemble forecasts. Q.R.J. Meteorol. Soc., 140:
      1917–1923, 2013. doi: 10.1002/qj.2270.
    * https://www-miklip.dkrz.de/about/problems/
    """
    with xr.set_options(keep_attrs=keep_attrs):
        if fair:
            if member_dim not in forecasts.dims:
                raise ValueError("need forecast with member dim")
            else:
                M = forecasts[member_dim].size
                e = (forecasts == 1).sum(member_dim)
                o = observations
                res = (e / M - o) ** 2 - e * (M - e) / (M ** 2 * (M - 1))
        else:
            if member_dim in forecasts.dims:
                forecasts = forecasts.mean(member_dim)
            res = xr.apply_ufunc(
                properscoring.brier_score,
                observations,
                forecasts,
                input_core_dims=[[], []],
                dask="parallelized",
                output_dtypes=[float],
                keep_attrs=keep_attrs,
            )
        if weights is not None:
            return res.weighted(weights).mean(dim)
        else:
            return res.mean(dim)


def threshold_brier_score(
    observations,
    forecasts,
    threshold,
    issorted=False,
    member_dim="member",
    dim=None,
    weights=None,
    keep_attrs=False,
):
    """Calculate the Brier scores of an ensemble for exceeding given thresholds.

    Parameters
    ----------
    observations : xarray.Dataset or xarray.DataArray
        The observations or set of observations.
    forecasts : xarray.Dataset or xarray.DataArray
        Forecast with required member dimension ``member_dim``.
    threshold : scalar or 1d scalar
        Threshold values at which to calculate exceedence Brier scores.
    issorted : bool, optional
        Optimization flag to indicate that the elements of `ensemble` are
        already sorted along `axis`.
    member_dim : str, optional
        Name of ensemble member dimension. By default, 'member'.
    dim : str or list of str, optional
        Dimension over which to compute mean after computing ``threshold_brier_score``.
        Defaults to None implying averaging over all dimensions.
    weights : xr.DataArray with dimensions from dim, optional
        Weights for `weighted.mean(dim)`. Defaults to None, such that no weighting is
        applied.
    keep_attrs : bool
        If True, the attributes (attrs) will be copied from the first input to the new
        one. If False (default), the new object will be returned without attributes.

    Returns
    -------
    xarray.Dataset or xarray.DataArray
        (If ``threshold`` is a scalar, the result will have the same shape as
        observations. Otherwise, it will have an additional final dimension
        corresponding to the threshold levels. Not implemented yet.)

    Examples
    --------
    >>> observations = xr.DataArray(np.random.normal(size=(3,3)),
    ...                             coords=[('x', np.arange(3)),
    ...                                     ('y', np.arange(3))])
    >>> forecasts = xr.DataArray(np.random.normal(size=(3,3,3)),
    ...                          coords=[('x', np.arange(3)),
    ...                                  ('y', np.arange(3)),
    ...                                  ('member', np.arange(3))])
    >>> threshold = [.2, .5, .8]
    >>> xs.threshold_brier_score(observations, forecasts, threshold)
    <xarray.DataArray (threshold: 3)>
    array([0.27160494, 0.34567901, 0.18518519])
    Coordinates:
      * threshold  (threshold) float64 0.2 0.5 0.8

    See Also
    --------
    properscoring.threshold_brier_score

    References
    ----------
    Gneiting, T. and Ranjan, R. Comparing density forecasts using threshold-
      and quantile-weighted scoring rules. J. Bus. Econ. Stat. 29, 411-422
      (2011). http://www.stat.washington.edu/research/reports/2008/tr533.pdf
    """
    if isinstance(threshold, list):
        threshold.sort()
        threshold = xr.DataArray(
            threshold, dims="threshold", coords={"threshold": threshold}
        )

    if isinstance(threshold, (xr.DataArray, xr.Dataset)):
        if "threshold" not in threshold.dims:
            raise ValueError(
                "please provide threshold as xr.DataArray with threshold dim, found",
                f"dim = {threshold.dims}, type = {type(threshold)}",
            )
        input_core_dims = [[], [member_dim], ["threshold"]]
        output_core_dims = [["threshold"]]
    elif isinstance(threshold, (int, float)):
        input_core_dims = [[], [member_dim], []]
        output_core_dims = [[]]
    else:
        raise ValueError(
            "Please provide threshold as list, int, float \
            or xr.object with threshold dimension; found",
            type(threshold),
        )
    res = xr.apply_ufunc(
        properscoring.threshold_brier_score,
        observations,
        forecasts,
        threshold,
        input_core_dims=input_core_dims,
        kwargs={"axis": -1, "issorted": issorted},
        output_core_dims=output_core_dims,
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=keep_attrs,
    )
    res = res.assign_coords(threshold=threshold)
    if dim is None:  # dont average over threshold dim
        dim = observations.dims
    if weights is not None:
        return res.weighted(weights).mean(dim, keep_attrs=keep_attrs)
    else:
        return res.mean(dim, keep_attrs=keep_attrs)


def _assign_rps_category_bounds(res, edges, name, bin_dim="category_edge"):
    """Add category_edge coord to rps return.
    Additionally adds left-most -np.inf category and right-most +np.inf category."""
    if edges[bin_dim].size >= 2:
        res = res.assign_coords(
            {
                f"{name}_category_edge": ", ".join(
                    _get_category_bounds(edges[bin_dim].values)
                )
            }
        )
        res[
            f"{name}_category_edge"
        ] = f"[-np.inf, {edges[bin_dim].isel({bin_dim:0}).values}), {str(res[f'{name}_category_edge'].values)[:-1]}), [{edges[bin_dim].isel({bin_dim:-1}).values}, np.inf]"
    return res


def rps(
    observations,
    forecasts,
    category_edges,
    dim=None,
    fair=False,
    weights=None,
    keep_attrs=False,
    member_dim="member",
    input_distributions=None,
):
    """Calculate Ranked Probability Score.

     .. math::
        RPS = \\sum_{m=1}^{M}[(\\sum_{k=1}^{m} y_k) - (\\sum_{k=1}^{m} o_k)]^{2}

    where ``y`` and ``o`` are forecast and observation probabilities in ``M``
    categories.

     .. note::
        Takes the sum over all categories as in Weigel et al. 2007 and not the mean as
        in https://www.cawcr.gov.au/projects/verification/verif_web_page.html#RPS.
        Therefore RPS has no upper boundary.

    Parameters
    ----------
    observations : xarray.Dataset or xarray.DataArray
        The observations of the event.
        Further requirements are specified based on ``category_edges``.
    forecasts : xarray.Dataset or xarray.DataArray
        The forecast of the event with dimension specified by ``member_dim``.
        Further requirements are specified based on ``category_edges``.
    category_edges : array_like, xr.Dataset, xr.DataArray, None
        If forecasts and observations are probabilistic, use ``category_edges=None``
        and set ``input_distributions``. If forecasts are deterministic and given in
        absolute units, set ``category_edges`` and leave ``input_distributions=None``.

        Edges (left-edge inclusive) of the bins used to calculate the cumulative
        density function (cdf). Note that here the bins have to include the full range
        of observations and forecasts data. Effectively, negative infinity is appended
        to the left side of category_edges, and positive infinity is appended to the
        right side. Thus, N category edges produces N+1 bins. For example, specifying
        category_edges = [0,1] will compute the RPS for bins [-inf, 0), [0, 1) and
        [1, inf), which results in CDF bins [-inf, 0), [-inf, 1) and [-inf, inf).
        Note that the edges are right-edge exclusive. Forecasts, observations and
        category_edge are expected in absolute units or probabilities consistently.
        dtypes are handled accordingly:

        - np.array (1d): will be internally converted and broadcasted to observations.
          Use this if you wish to use the same category edges for all elements of both
          forecasts and observations.

        - xr.Dataset/xr.DataArray: edges of the categories provided
          as dimension ``category_edge`` with optional category labels as
          ``category_edge`` coordinate. Use xr.Dataset/xr.DataArray if edges
          multi-dimensional and vary across dimensions. Use this if your category edges
          vary across dimensions of forecasts and observations, but are the same for
          both.

        - tuple of np.array/xr.Dataset/xr.DataArray: same as above, where the
          first item is taken as ``category_edges`` for observations and the second item
          for ``category_edges`` for forecasts. Use this if your category edges vary
          across dimensions of forecasts and observations, and are different for each.

        - None: expect that observations and forecasts are already cumulative
          distribution functions (``c``) or probability density functions (``p``), as
          specified by the ``input_distributions`` arg. In this case, the inputs
          observations and forecasts must contain a dimension named ``category`` with
          the bin values of the density function.

    dim : str or list of str, optional
        Dimension over which to mean after computing ``rps``. This represents a mean
        over multiple forecasts-observations pairs. Defaults to None implying averaging
        over all dimensions.
    fair: boolean
        Apply ensemble member-size adjustment for unbiased, fair metric; see Ferro
        (2013). If ``fair==True`` and ``category_edges==None``, require forecasts to
        have number of members in coordinates as forecasts[member_dim].
    weights : xr.DataArray with dimensions from dim, optional
        Weights for `weighted.mean(dim)`. Defaults to None, such that no weighting is
        applied.
    keep_attrs : bool
        If True, the attributes (attrs) will be copied from the first input to the new
        one. If False (default), the new object will be returned without attributes.
    member_dim : str, optional
        Name of ensemble member dimension. By default, 'member'.
    input_distributions: str or None
        Indicates whether observations and forecasts are probability distributions by
        ``p`` or cumulative distributions by ``c``.
        Only valid if `category_edges` is None. Defaults: None.

    Returns
    -------
    xarray.Dataset or xarray.DataArray:
        ranked probability score with coords ``forecasts_category_edge`` and
        ``observations_category_edge`` as str


    Examples
    --------
    In the examples below `o` is used for observations and `f` for forecasts.

    >>> o = xr.DataArray(np.random.random(size=(3, 3)),
    ...                  coords=[('x', np.arange(3)),
    ...                          ('y', np.arange(3))])
    >>> f = xr.DataArray(np.random.random(size=(3, 3, 3)),
    ...                  coords=[('x', np.arange(3)),
    ...                          ('y', np.arange(3)),
    ...                          ('member', np.arange(3))])
    >>> category_edges = np.array([.33, .66])
    >>> xs.rps(o, f, category_edges, dim='x')
    <xarray.DataArray (y: 3)>
    array([0.37037037, 0.81481481, 1.        ])
    Coordinates:
      * y                           (y) int64 0 1 2
        observations_category_edge  <U45 '[-np.inf, 0.33), [0.33, 0.66), [0.66, n...
        forecasts_category_edge     <U45 '[-np.inf, 0.33), [0.33, 0.66), [0.66, n...

    You can also define multi-dimensional ``category_edges``, e.g. with xr.quantile.
    However, you still need to ensure that ``category_edges`` covers the forecasts
    and observations distributions.

    >>> category_edges = o.quantile(q=[.33, .66]).rename({'quantile': 'category_edge'})
    >>> xs.rps(o, f, category_edges, dim='x')
    <xarray.DataArray (y: 3)>
    array([0.37037037, 0.81481481, 0.88888889])
    Coordinates:
      * y                           (y) int64 0 1 2
        observations_category_edge  <U45 '[-np.inf, 0.33), [0.33, 0.66), [0.66, n...
        forecasts_category_edge     <U45 '[-np.inf, 0.33), [0.33, 0.66), [0.66, n...

    If you have probabilistic forecasts, i.e. without a ``member`` dimension but
    different ``category`` probabilities, you can also provide probability
    distribution inputs by specifying ``category_edges=None`` and
    ``input_distributions``:

    >>> category_edges = category_edges.rename({'category_edge': 'category'})
    >>> categories = ["below normal", "normal", "above_normal"]
    >>> o_p = xr.concat([
    ...     (o < category_edges.isel(category=0)
    ...         ).assign_coords(category=categories[0]),
    ...     ((o >= category_edges.isel(category=0)) & (
    ...         o < category_edges.isel(category=1))
    ...         ).assign_coords(category=categories[1]),
    ...     (o >= category_edges.isel(category=1)
    ...         ).assign_coords(category=categories[2])
    ... ], 'category')
    >>> assert (o_p.sum('category')==1).all()
    >>> f_p = xr.concat([
    ...     (f < category_edges.isel(category=0)
    ...         ).assign_coords(category=categories[0]),
    ...     ((f >= category_edges.isel(category=0)
    ...         ) & (f < category_edges.isel(category=1))
    ...         ).assign_coords(category=categories[1]),
    ...     (f >= category_edges.isel(category=1)
    ...         ).assign_coords(category=categories[2])
    ... ], 'category').mean('member')
    >>> assert (f_p.sum('category')==1).all()
    >>> xs.rps(o_p, f_p, category_edges=None, dim='x', input_distributions='p')
    <xarray.DataArray (y: 3)>
    array([0.37037037, 0.81481481, 0.88888889])
    Coordinates:
      * y        (y) int64 0 1 2

    Providing cumulative distribution inputs yields identical results, where highest
    category equals 1 by default and can be ignored:

    >>> o_c = o < category_edges
    >>> f_c = (f < category_edges).mean('member')
    >>> xs.rps(o_c, f_c, category_edges=None, dim='x', input_distributions='c')
    <xarray.DataArray (y: 3)>
    array([0.37037037, 0.81481481, 0.88888889])
    Coordinates:
      * y        (y) int64 0 1 2

    References
    ----------
    * Weigel, A. P., Liniger, M. A., & Appenzeller, C. (2007). The Discrete Brier and
      Ranked Probability Skill Scores. Monthly Weather Review, 135(1), 118–124.
      doi: 10/b59qz5
    * C. A. T. Ferro. Fair scores for ensemble forecasts. Q.R.J. Meteorol. Soc., 140:
      1917–1923, 2013. doi: 10.1002/qj.2270.
    * https://www-miklip.dkrz.de/about/problems/
    """
    bin_dim = "category_edge"
    if category_edges is not None:
        if member_dim not in forecasts.dims:
            raise ValueError(
                f"Expect to find {member_dim} in forecasts dimensions, found"
                f"{forecasts.dims}."
            )
        if fair:
            M = forecasts[member_dim].size
    elif category_edges is None and fair:
        if member_dim in forecasts.coords and member_dim not in forecasts.dims:
            M = forecasts[member_dim].astype("int")
        else:
            raise ValueError(
                "category_edges=None and fair=True only works "
                "if forecast[member_dim] is a number in forecasts.coords."
            )

    forecasts = _bool_to_int(forecasts)

    _check_identical_xr_types(observations, forecasts)

    # different entry point of calculating RPS based on category_edges
    # category_edges tuple of two: use for obs and forecast category_edges separately
    if isinstance(category_edges, (tuple, np.ndarray, xr.DataArray, xr.Dataset)):
        if isinstance(category_edges, tuple):
            assert isinstance(category_edges[0], type(category_edges[1]))
            observations_edges, forecasts_edges = category_edges
        else:  # category_edges only given once, so use for both obs and forecasts
            observations_edges, forecasts_edges = category_edges, category_edges

        if isinstance(observations_edges, np.ndarray):
            # convert category_edges as xr object
            observations_edges = xr.DataArray(
                observations_edges,
                dims="category_edge",
                coords={"category_edge": observations_edges},
            )
            observations_edges = xr.ones_like(observations) * observations_edges

            forecasts_edges = xr.DataArray(
                forecasts_edges,
                dims="category_edge",
                coords={"category_edge": forecasts_edges},
            )
            forecasts_edges = (
                xr.ones_like(
                    forecasts
                    if member_dim not in forecasts.dims
                    else forecasts.isel({member_dim: 0}, drop=True)
                )
                * forecasts_edges
            )

        _check_identical_xr_types(forecasts_edges, forecasts)
        _check_identical_xr_types(observations_edges, forecasts)

        # cumulative probability functions
        # lowest category is [-np.inf, category_edges.isel(category_edge=0)]
        # ignores the right-most edge. The effective right-most edge is np.inf.
        # therefore the CDFs Fc and Oc both reach 1 for the right-most edge.
        # < makes edges right-edge exclusive
        Fc = (forecasts < forecasts_edges).mean(member_dim)
        Oc = (observations < observations_edges).astype("int")

    elif category_edges is None:  # expect inputs as cdfs or pdfs
        if input_distributions not in ["c", "p"]:
            raise ValueError(
                "If forecasts and observations are probabilistic, you use correctly"
                " ``category_edges==None``, but ``input_distributions`` must be"
                " either ``c`` for cumulative or ``p`` for probability"
                f" found ``input_distributions={input_distributions}``"
            )
        category_dim = "category"
        if category_dim not in forecasts.dims:
            raise ValueError(
                f"Expected dimension {category_dim} in cumulative forecasts, "
                f"found {forecasts.dims}"
            )
        if category_dim not in observations.dims:
            raise ValueError(
                f"Expected dimension {category_dim} in cumulative observations, "
                f"found {observations.dims}"
            )

        if member_dim in forecasts.dims:
            forecasts = forecasts.mean(member_dim)

        forecasts = forecasts.rename({category_dim: bin_dim})
        observations = observations.rename({category_dim: bin_dim})

        if input_distributions == "p":  # convert to cdf
            Fc = forecasts.cumsum(bin_dim)
            Oc = observations.cumsum(bin_dim)
        else:
            Fc = forecasts
            Oc = observations

    else:
        raise ValueError(
            "category_edges must be xr.DataArray, xr.Dataset, tuple of xr.objects, "
            f" None or array-like, found {type(category_edges)}"
        )

    # RPS formulas
    if fair:  # for ensemble member adjustment, see Ferro 2013
        Ec = Fc * M
        res = ((Ec / M - Oc) ** 2 - Ec * (M - Ec) / (M ** 2 * (M - 1))).sum(bin_dim)
    else:  # normal formula
        res = ((Fc - Oc) ** 2).sum(bin_dim)

    # add category_edge as str into coords
    if category_edges is not None:
        res = _assign_rps_category_bounds(res, observations_edges, "observations")
        res = _assign_rps_category_bounds(res, forecasts_edges, "forecasts")
    if weights is not None:
        res = res.weighted(weights)
    # combine many forecasts-observations pairs
    res = res.mean(dim)
    # keep nans and prevent 0 for all nan grids, could prevent this with skipna=False
    try:
        res = _keep_nans_masked(observations, res, dim, ignore=["category_edge"])
    except Exception as e:
        print(f"could not mask all NaNs properly due to {type(e).__name__}: {e}")
    if keep_attrs:  # attach by hand
        res.attrs.update(observations.attrs)
        res.attrs.update(forecasts.attrs)
        if isinstance(res, xr.Dataset):
            for v in res.data_vars:
                res[v].attrs.update(observations[v].attrs)
                res[v].attrs.update(forecasts[v].attrs)
    return res


def rank_histogram(observations, forecasts, dim=None, member_dim="member"):
    """Returns the rank histogram (Talagrand diagram) along the specified dimensions.

    Parameters
    ----------
    observations : xarray.Dataset or xarray.DataArray
        The observations or set of observations.
    forecasts : xarray.Dataset or xarray.DataArray
        Forecast with required member dimension ``member_dim``.
    dim : str or list of str, optional
        Dimension(s) over which to compute the histogram of ranks.
        Defaults to None meaning compute over all dimensions
    member_dim : str, optional
        Name of ensemble member dimension. By default, 'member'.

    Returns
    -------
    rank_histogram : xarray.Dataset or xarray.DataArray
        New object containing the histogram of ranks

    Examples
    --------
    >>> observations = xr.DataArray(np.random.normal(size=(3, 3)),
    ...                             coords=[('x', np.arange(3)),
    ...                                     ('y', np.arange(3))])
    >>> forecasts = xr.DataArray(np.random.normal(size=(3, 3, 3)),
    ...                          coords=[('x', np.arange(3)),
    ...                                  ('y', np.arange(3)),
    ...                                  ('member', np.arange(3))])
    >>> xs.rank_histogram(observations, forecasts, dim='x')
    <xarray.DataArray 'histogram_rank' (y: 3, rank: 4)>
    array([[0, 0, 1, 2],
           [0, 1, 2, 0],
           [0, 0, 2, 1]])
    Coordinates:
      * y        (y) int64 0 1 2
      * rank     (rank) float64 1.0 2.0 3.0 4.0

    Notes
    -----
    See http://www.cawcr.gov.au/projects/verification/
    """

    def _rank_first(x, y):
        """Concatenates x and y and returns the rank of the
        first element along the last axes"""
        xy = np.concatenate((x[..., np.newaxis], y), axis=-1)
        return bn.nanrankdata(xy, axis=-1)[..., 0]

    if dim is not None:
        if len(dim) == 0:
            raise ValueError(
                "At least one dimension must be supplied to compute rank histogram over"
            )
        if member_dim in dim:
            raise ValueError(f'"{member_dim}" cannot be specified as an input to dim')

    ranks = xr.apply_ufunc(
        _rank_first,
        observations,
        forecasts,
        input_core_dims=[[], [member_dim]],
        dask="parallelized",
        output_dtypes=[int],
    )

    bin_edges = np.arange(0.5, len(forecasts[member_dim]) + 2)
    return histogram(
        ranks, bins=[bin_edges], bin_names=["rank"], dim=dim, bin_dim_suffix=""
    )


def discrimination(
    observations,
    forecasts,
    dim=None,
    probability_bin_edges=np.linspace(0, 1, 6),
):
    """Returns the data required to construct the discrimination diagram for an event;
       the histogram of forecasts likelihood when observations indicate an event has
       occurred and has not occurred.

    Parameters
    ----------
    observations : xarray.Dataset or xarray.DataArray
        The observations or set of observations of the event.
        Data should be boolean or logical \
        (True or 1 for event occurance, False or 0 for non-occurance).
    forecasts : xarray.Dataset or xarray.DataArray
        The forecast likelihoods of the event. Data should be between 0 and 1.
    dim : str or list of str, optional
        Dimension(s) over which to compute the histograms
        Defaults to None meaning compute over all dimensions.
    probability_bin_edges : array_like, optional
        Probability bin edges used to compute the histograms. Similar to np.histogram, \
        all but the last (righthand-most) bin include the left edge and exclude the \
        right edge. The last bin includes both edges. Defaults to 6 equally spaced \
        edges between 0 and 1

    Returns
    -------
    xarray.Dataset or xarray.DataArray
        Array with added dimension "event" containing the histograms of
        forecast probabilities when the event was observed and not observed

    Examples
    --------
    >>> observations = xr.DataArray(np.random.normal(size=(30, 30)),
    ...                             coords=[('x', np.arange(30)),
    ...                                     ('y', np.arange(30))])
    >>> forecasts = xr.DataArray(np.random.normal(size=(30, 30, 10)),
    ...                          coords=[('x', np.arange(30)),
    ...                                  ('y', np.arange(30)),
    ...                                  ('member', np.arange(10))])
    >>> forecast_event_likelihood = (forecasts > 0).mean('member')
    >>> observed_event = observations > 0
    >>> xs.discrimination(observed_event, forecast_event_likelihood, dim=['x','y'])
    <xarray.DataArray (event: 2, forecast_probability: 5)>
    array([[0.00437637, 0.15536105, 0.66739606, 0.12472648, 0.04814004],
           [0.00451467, 0.16704289, 0.66365688, 0.1241535 , 0.04063205]])
    Coordinates:
      * forecast_probability  (forecast_probability) float64 0.1 0.3 0.5 0.7 0.9
      * event                 (event) bool True False

    References
    ----------
    http://www.cawcr.gov.au/projects/verification/
    """

    _fail_if_dim_empty(dim)

    hist_event = (
        histogram(
            forecasts.where(observations),
            bins=[probability_bin_edges],
            bin_names=[FORECAST_PROBABILITY_DIM],
            bin_dim_suffix="",
            dim=dim,
        )
        / (observations).sum(dim=dim)
    )

    hist_no_event = (
        histogram(
            forecasts.where(np.logical_not(observations)),
            bins=[probability_bin_edges],
            bin_names=[FORECAST_PROBABILITY_DIM],
            bin_dim_suffix="",
            dim=dim,
        )
        / (np.logical_not(observations)).sum(dim=dim)
    )

    return xr.concat([hist_event, hist_no_event], dim="event").assign_coords(
        {"event": [True, False]}
    )


def reliability(
    observations,
    forecasts,
    dim=None,
    probability_bin_edges=np.linspace(0, 1, 6),
    keep_attrs=False,
):
    """Returns the data required to construct the reliability diagram for an event;
        the relative frequencies of occurrence of an event
        for a range of forecast probability bins

    Parameters
    ----------
    observations : xarray.Dataset or xarray.DataArray
        The observations or set of observations of the event.
        Data should be boolean or logical \
        (True or 1 for event occurance, False or 0 for non-occurance).
    forecasts : xarray.Dataset or xarray.DataArray
        The forecast likelihoods of the event. Data should be between 0 and 1.
    dim : str or list of str, optional
        Dimension(s) over which to compute the histograms
        Defaults to None meaning compute over all dimensions.
    probability_bin_edges : array_like, optional
        Probability bin edges used to compute the reliability. Similar to np.histogram,
        all but the last (righthand-most) bin include the left edge and exclude the \
        right edge. The last bin includes both edges. Defaults to 6 equally spaced \
        edges between 0 and 1
    keep_attrs : bool, optional
        If True, the attributes (attrs) will be copied from the first input to
        the new one.
        If False (default), the new object will be returned without attributes.

    Returns
    -------
    xarray.Dataset or xarray.DataArray
        The relative frequency of occurrence for each probability bin

    Examples
    --------
    >>> forecasts = xr.DataArray(np.random.normal(size=(3,3,3)),
    ...                          coords=[('x', np.arange(3)),
    ...                                  ('y', np.arange(3)),
    ...                                  ('member', np.arange(3))])
    >>> observations = xr.DataArray(np.random.normal(size=(3,3)),
    ...                             coords=[('x', np.arange(3)),
    ...                                     ('y', np.arange(3))])
    >>> xs.reliability(observations > 0.1,
    ...                (forecasts > 0.1).mean('member'),
    ...                dim='x')
    <xarray.DataArray (y: 3, forecast_probability: 5)>
    array([[nan, 0. , nan, 1. , nan],
           [1. , 0.5, nan, nan, nan],
           [nan, 0. , nan, 0. , nan]])
    Coordinates:
      * y                     (y) int64 0 1 2
      * forecast_probability  (forecast_probability) float64 0.1 0.3 0.5 0.7 0.9
        samples               (y, forecast_probability) float64 0.0 2.0 ... 1.0 0.0

    Notes
    -----
    See http://www.cawcr.gov.au/projects/verification/
    """

    def _reliability(o, f, bin_edges):
        """Return the reliability and number of samples per bin"""
        # I couldn't get dask='parallelized' working in this case
        # so dealing with dask arrays explicitly
        is_dask_array = isinstance(o, darray.core.Array) | isinstance(
            f, darray.core.Array
        )

        if is_dask_array:
            r = []
            N = []
        else:
            r = np.zeros((*o.shape[:-1], len(bin_edges) - 1), dtype=float)
            N = np.zeros_like(r)

        for i in range(len(bin_edges) - 1):
            # Follow xhistogram behaviour: all bins are half-open, except for the right-most bin
            # which adds an epsilon to the right edge
            # see https://github.com/xgcm/xhistogram/issues/18
            if i == (len(bin_edges) - 2):
                f_in_bin = (f >= bin_edges[i]) & (f < (bin_edges[i + 1] + 1e-8))
            else:
                f_in_bin = (f >= bin_edges[i]) & (f < bin_edges[i + 1])
            o_f_in_bin = o & f_in_bin
            N_f_in_bin = f_in_bin.sum(axis=-1)
            N_o_f_in_bin = o_f_in_bin.sum(axis=-1)
            if is_dask_array:
                r.append(N_o_f_in_bin / N_f_in_bin)
                N.append(N_f_in_bin)
            else:
                with suppress_warnings("invalid value encountered in true_divide"):
                    with suppress_warnings("invalid value encountered in long_scalars"):
                        r[..., i] = N_o_f_in_bin / N_f_in_bin
                        N[..., i] = N_f_in_bin

        if is_dask_array:
            return (
                darray.stack(r, axis=-1).rechunk({-1: -1}),
                darray.stack(N, axis=-1).rechunk({-1: -1}),
            )
        else:
            return r, N

    _fail_if_dim_empty(dim)

    # Compute over all dims if dim is None
    if dim is None:
        dim = list(observations.dims)

    dim, _ = _preprocess_dims(dim, observations)
    observations, forecasts, stack_dim, _ = _stack_input_if_needed(
        observations, forecasts, dim, weights=None
    )

    rel, samp = xr.apply_ufunc(
        _reliability,
        observations,
        forecasts,
        probability_bin_edges,
        input_core_dims=[[stack_dim], [stack_dim], []],
        dask="allowed",
        output_core_dims=[[FORECAST_PROBABILITY_DIM], [FORECAST_PROBABILITY_DIM]],
        keep_attrs=keep_attrs,
    )

    # Add probability bin coordinate
    rel = rel.assign_coords(
        {FORECAST_PROBABILITY_DIM: _get_bin_centers(probability_bin_edges)}
    )
    samp = samp.assign_coords(
        {FORECAST_PROBABILITY_DIM: _get_bin_centers(probability_bin_edges)}
    )

    # Move samples to a coordinate
    return _add_as_coord(rel, samp, coordinate_suffix="samples")


def _drop_intermediate(fpr, tpr):
    """Attempt to drop thresholds corresponding to points in between and
    collinear with other points. These are always suboptimal and do not
    appear on a plotted ROC curve (and thus do not affect the AUC).
    Here xr.diff(_, 2) is used as a "second derivative" to tell if there is a
    corner at the point. (...) This keeps all cases where the point should be
    kept, but does not drop more complicated cases like fps = [1, 3, 7],
    # tps = [1, 2, 4]; there is no harm in keeping too many thresholds.
    https://github.com/scikit-learn/scikit-learn/blob/42aff4e2edd8e8887478f6ff1628f27de97be6a3/sklearn/metrics/_ranking.py#L916
    """
    optimal_idxs = xr.concat(
        [
            fpr.isel(probability_bin=0, drop=False).astype("bool"),
            np.logical_or(
                fpr.diff("probability_bin", 2), tpr.diff("probability_bin", 2)
            ),
            fpr.isel(probability_bin=0, drop=False).astype("bool"),
        ],
        "probability_bin",
    )
    optimal_idxs["probability_bin"] = np.arange(optimal_idxs.probability_bin.size)
    if isinstance(optimal_idxs, xr.Dataset):
        optimal_idxs = optimal_idxs.to_array()
    with suppress_warnings("invalid value encountered in true_divide"):
        with suppress_warnings("invalid value encountered in long_scalars"):
            optimal_idxs = optimal_idxs.where(
                optimal_idxs, drop=True
            ).probability_bin.values
    tpr = tpr.isel(probability_bin=optimal_idxs)
    fpr = fpr.isel(probability_bin=optimal_idxs)
    return fpr, tpr


def _auc(fpr, tpr, dim="probability_bin"):
    """Get area under the curve with trapez method."""
    # reverse tpr, fpr to fpr, tpr, see numpy.trapz(y, x=None)
    with suppress_warnings("The `numpy.trapz` function is not implemented"):
        with suppress_warnings("invalid value encountered in long_scalars"):
            with suppress_warnings("invalid value encountered in true_divide"):
                area = xr.apply_ufunc(
                    np.trapz, tpr, fpr, input_core_dims=[[dim], [dim]], dask="allowed"
                )
    area = abs(area)
    if ((area > 1)).any():
        area = area.clip(0, 1)  # allow only values between 0 and 1
    return area


def roc(
    observations,
    forecasts,
    bin_edges="continuous",
    dim=None,
    drop_intermediate=False,
    return_results="area",
):
    """Computes the relative operating characteristic for a range of thresholds.

    Parameters
    ----------
    observations : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
        If ``bin_edges=='continuous'``, observations are binary.
    forecasts : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
        If ``bin_edges=='continuous'``, forecasts are probabilities.
    bin_edges : array_like, str, default='continuous'
        Bin edges for categorising observations and forecasts. Similar to np.histogram, \
        all but the last (righthand-most) bin include the left edge and exclude the \
        right edge. The last bin includes both edges. ``bin_edges`` will be sorted in \
        ascending order. If ``bin_edges=='continuous'``, calculate ``bin_edges`` from \
        forecasts, equal to ``sklearn.metrics.roc_curve(f_boolean, o_prob)``.
    dim : str, list
        The dimension(s) over which to compute the contingency table
    drop_intermediate : bool, default=False
        Whether to drop some suboptimal thresholds which would not appear on a plotted
        ROC curve. This is useful in order to create lighter ROC curves.
        Defaults to ``True`` in ``sklearn.metrics.roc_curve``.
    return_results: str, default='area'
        Specify how return is structed:

            - 'area': return only the ``area under curve`` of ROC

            - 'all_as_tuple': return ``true positive rate`` and ``false positive rate``
              at each bin and area under the curve of ROC as tuple

            - 'all_as_metric_dim': return ``true positive rate`` and
              ``false positive rate`` at each bin and ``area under curve`` of ROC
              concatinated into new ``metric`` dimension

    Returns
    -------
    xarray.Dataset or xarray.DataArray :
        reduced by dimensions ``dim``, see ``return_results`` parameter.
        ``true positive rate`` and ``false positive rate`` contain
        ``probability_bin`` dimension with ascending ``bin_edges`` as coordinates.

    Examples
    --------
    >>> f = xr.DataArray(np.random.normal(size=(1000)),
    ...                  coords=[('time', np.arange(1000))])
    >>> o = f.copy()
    >>> category_edges = np.linspace(-2, 2, 5)
    >>> xs.roc(o, f, category_edges, dim=['time'])
    <xarray.DataArray 'histogram_observations_forecasts' ()>
    array(1.)

    See also
    --------
    xskillscore.Contingency
    sklearn.metrics.roc_curve

    References
    ----------
    http://www.cawcr.gov.au/projects/verification/
    """

    if dim is None:
        dim = list(forecasts.dims)
    if isinstance(dim, str):
        dim = [dim]

    continuous = False
    if isinstance(bin_edges, str):
        if bin_edges == "continuous":
            continuous = True
            # check that o binary
            if isinstance(observations, xr.Dataset):
                o_check = observations.to_array()
            else:
                o_check = observations
            if str(o_check.dtype) != "bool":
                if not ((o_check == 0) | (o_check == 1)).all():
                    raise ValueError(
                        'Input "observations" must represent logical (True/False) outcomes',
                        o_check,
                    )

            # works only for 1var
            if isinstance(forecasts, xr.Dataset):
                varlist = list(forecasts.data_vars)
                if len(varlist) == 1:
                    v = varlist[0]
                else:
                    raise ValueError(
                        "Only works for `xr.Dataset` with one variable, found"
                        f"{forecasts.data_vars}. Considering looping over `data_vars`"
                        "or `.to_array()`."
                    )
                f_bin = forecasts[v]
            else:
                f_bin = forecasts
            f_bin = f_bin.stack(ndim=forecasts.dims)
            f_bin = f_bin.sortby(-f_bin)
            bin_edges = np.append(f_bin[0] + 1, f_bin)
            bin_edges = np.unique(bin_edges)  # ensure that in ascending order
        else:
            raise ValueError("If bin_edges is str, it can only be continuous.")
    else:
        bin_edges = np.sort(bin_edges)  # ensure that in ascending order

    # loop over each bin_edge and get true positive rate and false positive rate
    # from contingency
    tpr, fpr = [], []
    for i in bin_edges:
        dichotomous_category_edges = np.array(
            [-np.inf, i, np.inf]
        )  # "dichotomous" means two-category
        dichotomous_contingency = Contingency(
            observations,
            forecasts,
            dichotomous_category_edges,
            dichotomous_category_edges,
            dim=dim,
        )
        fpr.append(dichotomous_contingency.false_alarm_rate())
        tpr.append(dichotomous_contingency.hit_rate())
    tpr = xr.concat(tpr, "probability_bin")
    fpr = xr.concat(fpr, "probability_bin")
    tpr["probability_bin"] = bin_edges
    fpr["probability_bin"] = bin_edges

    fpr = fpr.fillna(1.0)
    tpr = tpr.fillna(0.0)

    # pad (0,0) and (1,1)
    fpr_pad = xr.concat(
        [
            xr.ones_like(fpr.isel(probability_bin=0, drop=False)),
            fpr,
            xr.zeros_like(fpr.isel(probability_bin=-1, drop=False)),
        ],
        "probability_bin",
    )
    tpr_pad = xr.concat(
        [
            xr.ones_like(tpr.isel(probability_bin=0, drop=False)),
            tpr,
            xr.zeros_like(tpr.isel(probability_bin=-1, drop=False)),
        ],
        "probability_bin",
    )

    if drop_intermediate and fpr.probability_bin.size > 2:

        fpr, tpr = _drop_intermediate(fpr, tpr)
        fpr_pad, tpr_pad = _drop_intermediate(fpr_pad, tpr_pad)

    area = _auc(fpr_pad, tpr_pad)

    if continuous:
        # sklearn returns in reversed order
        fpr = fpr.sortby(-fpr.probability_bin)
        tpr = tpr.sortby(-fpr.probability_bin)

    # mask always nan
    def _keep_masked(new, ori, dim):
        """Keep mask from `ori` deprived of dimensions from `dim` in input `new`."""
        isel_dim = {d: 0 for d in forecasts.dims if d in dim}
        mask = ori.isel(isel_dim, drop=True)
        new_masked = new.where(mask.notnull())
        return new_masked

    fpr = _keep_masked(fpr, forecasts, dim=dim)
    tpr = _keep_masked(tpr, forecasts, dim=dim)
    area = _keep_masked(area, forecasts, dim=dim)

    if return_results == "area":
        return area
    elif return_results == "all_as_metric_dim":
        results = xr.concat([fpr, tpr, area], "metric", coords="minimal")
        results["metric"] = [
            "false positive rate",
            "true positive rate",
            "area under curve",
        ]
        return results
    elif return_results == "all_as_tuple":
        return fpr, tpr, area
    else:
        raise NotImplementedError(
            "expect `return_results` from [all_as_tuple, area, all_as_metric_dim], "
            f"found {return_results}"
        )



import numpy as np
import pytest
import xarray as xr
from dask import is_dask_collection

from xskillscore.core.resampling import (
    _gen_idx,
    resample_iterations,
    resample_iterations_idx,
)

resample_iterations_funcs = [resample_iterations, resample_iterations_idx]
ITERATIONS = 2


def assert_dim_coords(a, b):
    """check dim and coord entries and order."""
    # check dim and coord size and entries
    for d in a.dims:
        assert a[d].size == b[d].size, print(a[d], "!=", b[d])
    for c in a.coords:
        assert (a[c] == b[c]).all(), print(a[c], "!=", b[c])
    # check dim and coords entries and order
    assert list(a.dims) == list(b.dims), print(a.dims, "!=", b.dims)
    assert list(a.coords) == list(b.coords), print(a.coords, "!=", b.coords)


def test_resampling_identical_dim(f_prob):
    """check that resampling functions have the same ordering of dims and coords."""
    da = f_prob
    r1 = resample_iterations(da, iterations=ITERATIONS)
    r2 = resample_iterations_idx(da, iterations=ITERATIONS)
    assert_dim_coords(r1, r2)


def test_resampling_roughly_identical_mean(f_prob):
    """check that resampling functions result in allclose iteration mean."""
    da = f_prob.isel(lon=0, lat=0, drop=True)
    iterations = 1000
    r1 = resample_iterations(da, iterations=iterations, replace=True)
    r2 = resample_iterations_idx(da, iterations=iterations, replace=True)
    xr.testing.assert_allclose(
        r1.mean("iteration"), r2.mean("iteration"), atol=0.05, rtol=0.05
    )


@pytest.mark.parametrize("iterations", [1, 2])
@pytest.mark.parametrize("func", resample_iterations_funcs)
def test_resampling_same_dim_coord_order_as_input(func, f_prob, iterations):
    """check whether resampling function maintain dim and coord order and add iteration
    dimension to the end."""
    da = f_prob
    r1 = resample_iterations(da, iterations=iterations)
    # same as without having iteration
    assert_dim_coords(r1.isel(iteration=0, drop=True), da)
    # same as with having iteration added in the end as in expand_dims
    assert_dim_coords(
        r1,
        da.expand_dims("iteration")
        .transpose(..., "iteration")
        .isel(iteration=[0] * iterations)
        .assign_coords(iteration=range(iterations)),
    )


@pytest.mark.skip(reason="fails, dont know why and slow")
@pytest.mark.parametrize("func", resample_iterations_funcs)
def test_resampling_replace_True_larger_std_than_replace_False(f_prob, func):
    """check that resampling functions result in allclose iteration mean."""
    da = f_prob.isel(lon=0, lat=0, member=0, drop=True)
    iterations = 10000
    # I would hope to get the same random seed for both but probably thats unrealistic
    np.random.seed(42)
    rreplace = func(da, dim="time", iterations=iterations, replace=True)
    rnoreplace = func(da, dim="time", iterations=iterations, replace=False)
    assert (rreplace.std("iteration") > rnoreplace.std("iteration")).all(), print(
        rreplace.std("iteration"),
        "\nnot greater\n",
        rnoreplace.std("iteration"),
        "\nbut should\n",
        rreplace.std("iteration") > rnoreplace.std("iteration"),
    )


@pytest.mark.parametrize("replace", [True, False])
def test_gen_idx_replace(f_prob, replace):
    """check whether replace=True creates duplicate idx."""

    def find_duplicate(da, dim="iteration"):
        for i in da[dim]:
            if da.sel({dim: i}).to_index().duplicated().any():
                return True
        return False

    actual = _gen_idx(
        f_prob, "member", 10, f_prob["member"].size, replace, f_prob["member"]
    )
    if replace:  # find identical values
        assert find_duplicate(actual)
    else:  # find no identicals
        assert not find_duplicate(actual)


@pytest.mark.skip(reason="this is a bug, test fails and should be resolved.")
def test_resample_iterations_dix_no_squeeze(f_prob):
    """Test _resample_iteration_idx with singular dimension.

    Currently this fails for dimensions with just a single index as we use `squeeze` in
    the code and not using squeeze doesnt maintain functionality. This means that
    ``_resample_iteration_idx`` should not be called on singleton dimension inputs
    (which is not critical and can be circumvented when using squeeze before.).
    """
    da = f_prob.expand_dims("test_dim")
    actual = resample_iterations_idx(da, iterations=ITERATIONS)
    assert "test_dim" in actual.dims


@pytest.mark.parametrize("func", resample_iterations_funcs)
def test_resample_replace_False_once_same_mean(f_prob, func):
    """resample replace=False on one dimension once gets same mean as origin."""
    da = f_prob.isel(lon=0, lat=0, member=0, drop=True)
    once = func(da, dim="time", replace=False, iterations=1).squeeze(drop=True)
    xr.testing.assert_allclose(once.mean("time"), da.mean("time"))


@pytest.mark.parametrize("func", resample_iterations_funcs)
@pytest.mark.parametrize("dim_max", [None, 5])
def test_resample_dim_max(f_prob, func, dim_max):
    """resample dim_max=x returns only x items in dim."""
    da = f_prob.isel(lon=0, lat=0, member=0, drop=True)
    dim = "time"
    actual = func(da, dim=dim, dim_max=dim_max, iterations=ITERATIONS)
    if dim_max:
        assert actual[dim].size == dim_max
    else:
        assert actual[dim].size == da[dim].size


@pytest.mark.parametrize("replace", [True, False])
@pytest.mark.parametrize("chunk", [True, False])
@pytest.mark.parametrize("input", ["Dataset", "multidim Dataset", "DataArray"])
@pytest.mark.parametrize("func", resample_iterations_funcs)
def test_resample_inputs(a_1d, func, input, chunk, replace):
    """Test sign_test with xr inputs and chunked."""
    if "Dataset" in input:
        name = "var"
        a_1d = a_1d.to_dataset(name=name)
        if input == "multidim Dataset":
            a_1d["var2"] = a_1d["var"] * 2
    if chunk:
        a_1d = a_1d.chunk()
    actual = func(a_1d, ITERATIONS, dim="time")
    # check dask collection preserved
    assert is_dask_collection(actual) if chunk else not is_dask_collection(actual)
    # input type preserved
    assert type(actual) == type(a_1d)



# Write the benchmarking functions here.
# See "Writing benchmarks" in the asv docs for more information.


import numpy as np
import xarray as xr
from scipy.stats import norm

from xskillscore import (
    brier_score,
    crps_ensemble,
    crps_gaussian,
    crps_quadrature,
    threshold_brier_score,
)

from . import Generate, parameterized, requires_dask

PROBABILISTIC_METRICS = [
    crps_ensemble,
    crps_gaussian,
    crps_quadrature,
    brier_score,
    threshold_brier_score,
]

including_crps_quadrature = False

large_lon_lat = 2000
large_lon_lat_chunksize = large_lon_lat // 2
nmember = 4


class Compute_small(Generate):
    """
    A  benchmark xskillscore.metric for small xr.DataArrays"""

    def setup(self, *args, **kwargs):
        self.make_ds(nmember, 1, 1)  # no grid
        self.ds["tos"] = self.ds["tos"].isel(member=0, drop=True)

    @parameterized("metric", PROBABILISTIC_METRICS)
    def time_xskillscore_metric(self, metric):
        """Take time for xskillscore.metric."""
        if metric is crps_gaussian:
            mu = 0.5
            sig = 0.2
            metric(self.ds["tos"], mu, sig).compute()
        elif metric is crps_quadrature:
            if not including_crps_quadrature:
                pass
            else:
                xmin, xmax, tol = -10, 10, 1e-6
                cdf_or_dist = norm
                metric(self.ds["tos"], cdf_or_dist, xmin, xmax, tol).compute()
        elif metric is crps_ensemble:
            metric(self.ds["tos"], self.ds["sos"]).compute()
        elif metric is threshold_brier_score:
            threshold = 0.5
            metric(self.ds["tos"], self.ds["sos"], threshold).compute()
        elif metric is brier_score:
            metric(
                self.ds["tos"] > 0.5, (self.ds["sos"] > 0.5).mean("member")
            ).compute()

    @parameterized("metric", PROBABILISTIC_METRICS)
    def peakmem_xskillscore_metric(self, metric):
        """Take time for xskillscore.metric."""
        if metric is crps_gaussian:
            mu = 0.5
            sig = 0.2
            metric(self.ds["tos"], mu, sig).compute()
        elif metric is crps_quadrature:
            if not including_crps_quadrature:
                pass
            else:
                xmin, xmax, tol = -10, 10, 1e-6
                cdf_or_dist = norm
                metric(self.ds["tos"], cdf_or_dist, xmin, xmax, tol).compute()
        elif metric is crps_ensemble:
            metric(self.ds["tos"], self.ds["sos"]).compute()
        elif metric is threshold_brier_score:
            threshold = 0.5
            metric(self.ds["tos"], self.ds["sos"], threshold).compute()
        elif metric is brier_score:
            metric(
                self.ds["tos"] > 0.5, (self.ds["sos"] > 0.5).mean("member")
            ).compute()


class Compute_large(Compute_small):
    """
    A benchmark xskillscore.metric for large xr.DataArrays."""

    def setup(self, *args, **kwargs):
        self.make_ds(nmember, large_lon_lat, large_lon_lat)
        self.ds["tos"] = self.ds["tos"].isel(member=0, drop=True)


class Compute_large_dask(Compute_small):
    """
    A benchmark xskillscore.metric for large xr.DataArrays with dask."""

    def setup(self, *args, **kwargs):
        requires_dask()
        self.make_ds(
            nmember,
            large_lon_lat,
            large_lon_lat,
            chunks={"lon": large_lon_lat_chunksize},
        )
        self.ds["tos"] = self.ds["tos"].isel(member=0, drop=True)



import numpy as np
import numpy.testing as npt
import properscoring
import pytest
import xarray as xr
from dask import is_dask_collection
from pytest_lazyfixture import lazy_fixture
from scipy.stats import norm
from sklearn.calibration import calibration_curve
from sklearn.metrics import roc_auc_score, roc_curve
from xarray.tests import assert_allclose, assert_identical

from xskillscore.core.probabilistic import (
    brier_score,
    crps_ensemble,
    crps_gaussian,
    crps_quadrature,
    discrimination,
    rank_histogram,
    reliability,
    roc,
    rps,
    threshold_brier_score,
)
from xskillscore.core.utils import suppress_warnings

DIMS = ["lon", "lat", ["lon", "lat"], None, []]


def modify_inputs(o, f_prob, input_type, chunk_bool):
    """Modify inputs depending on input_type and chunk_bool."""
    if "Dataset" in input_type:
        name = "var"
        o = o.to_dataset(name=name)
        f_prob = f_prob.to_dataset(name=name)
        if input_type == "multidim Dataset":
            o["var2"] = o["var"] ** 2
            f_prob["var2"] = f_prob["var"] ** 2
    if chunk_bool:
        o = o.chunk()
        f_prob = f_prob.chunk()
    return o, f_prob


def assert_only_dim_reduced(dim, actual, obs):
    """Check that actual is only reduced by dims in dim."""
    if dim is None:
        dim = list(obs.dims)
    elif not isinstance(dim, list):
        dim = [dim]
    for d in dim:
        assert d not in actual.dims
    for d in obs.dims:
        if d not in dim:
            assert d in actual.dims


def assert_chunk(actual, chunk_bool):
    """check that actual is chunked when chunk_bool==True."""
    if chunk_bool:
        assert is_dask_collection(actual)
    else:
        assert not is_dask_collection(actual)


def assert_keep_attrs(actual, o, keep_attrs):
    """check that actual kept attributes only if keep_attrs==True."""
    if keep_attrs:
        assert actual.attrs == o.attrs
    else:
        assert actual.attrs == {}


def assign_type_input_output(actual, o):
    assert type(o) == type(actual)


@pytest.mark.parametrize("chunk_bool", [True, False])
@pytest.mark.parametrize("input_type", ["Dataset", "multidim Dataset", "DataArray"])
@pytest.mark.parametrize("keep_attrs", [True, False])
def test_crps_ensemble_api_and_inputs(o, f_prob, keep_attrs, input_type, chunk_bool):
    """Test that crps_ensemble keeps attributes, chunking, input types and equals
    properscoring.crps_ensemble."""
    o, f_prob = modify_inputs(o, f_prob, input_type, chunk_bool)
    actual = crps_ensemble(o, f_prob, keep_attrs=keep_attrs)
    if input_type == "DataArray":  # properscoring allows only DataArrays
        expected = properscoring.crps_ensemble(o, f_prob, axis=0)
        expected = xr.DataArray(expected, coords=o.coords).mean()
        # test for numerical identity of xskillscore crps and properscoring crps
        assert_allclose(actual, expected)
    # test that returns chunks
    assert_chunk(actual, chunk_bool)
    # test that attributes are kept
    assert_keep_attrs(actual, o, keep_attrs)
    # test that input types equal output types
    assign_type_input_output(actual, o)


@pytest.mark.parametrize("dim", DIMS)
def test_crps_ensemble_dim(o, f_prob, dim):
    """Check that crps_ensemble reduces only dim."""
    actual = crps_ensemble(o, f_prob, dim=dim)
    assert_only_dim_reduced(dim, actual, o)


def test_crps_ensemble_weighted(o, f_prob, weights_cos_lat):
    """Test that weighted changes results of crps_ensemble."""
    dim = ["lon", "lat"]
    actual = crps_ensemble(o, f_prob, dim=dim, weights=weights_cos_lat)
    actual_no_weights = crps_ensemble(o, f_prob, dim=dim)
    assert not (actual_no_weights == actual).all()


@pytest.mark.parametrize("chunk_bool", [True, False])
@pytest.mark.parametrize("input_type", ["Dataset", "multidim Dataset", "DataArray"])
@pytest.mark.parametrize("keep_attrs", [True, False])
def test_crps_gaussian_api_and_inputs(o, f_prob, keep_attrs, input_type, chunk_bool):
    """Test that crps_gaussian keeps attributes, chunking, input types and equals
    properscoring.crps_gaussian."""
    o, f_prob = modify_inputs(o, f_prob, input_type, chunk_bool)
    mu = f_prob.mean("member")
    sig = f_prob.std("member")
    actual = crps_gaussian(o, mu, sig, keep_attrs=keep_attrs)
    if input_type == "DataArray":  # properscoring allows only DataArrays
        expected = properscoring.crps_gaussian(o, mu, sig)
        expected = xr.DataArray(expected, coords=o.coords).mean()
        # test for numerical identity of xskillscore crps and properscoring crps
        assert_allclose(actual, expected)
    # test that returns chunks
    assert_chunk(actual, chunk_bool)
    # test that attributes are kept
    assert_keep_attrs(actual, o, keep_attrs)
    # test that input types equal output types
    assign_type_input_output(actual, o)


@pytest.mark.parametrize("dim", DIMS)
def test_crps_gaussian_dim(o, f_prob, dim):
    """Check that crps_gaussian reduces only dim."""
    mu = f_prob.mean("member")
    sig = f_prob.std("member")
    actual = crps_gaussian(o, mu, sig, dim=dim)
    assert_only_dim_reduced(dim, actual, o)


@pytest.mark.slow
@pytest.mark.parametrize("chunk_bool", [True, False])
@pytest.mark.parametrize("input_type", ["Dataset", "multidim Dataset", "DataArray"])
@pytest.mark.parametrize("keep_attrs", [True, False])
def test_crps_quadrature_api_and_inputs(o, f_prob, keep_attrs, input_type, chunk_bool):
    """Test that crps_quadrature keeps attributes, chunking, input types and equals
    properscoring.crps_quadrature."""
    o, f_prob = modify_inputs(o, f_prob, input_type, chunk_bool)
    # to speed things up
    o = o.isel(time=0, drop=True)
    cdf_or_dist = norm
    actual = crps_quadrature(o, cdf_or_dist, keep_attrs=keep_attrs)
    if input_type == "DataArray":  # properscoring allows only DataArrays
        expected = properscoring.crps_quadrature(o, cdf_or_dist)
        expected = xr.DataArray(expected, coords=o.coords).mean()
        # test for numerical identity of xskillscore crps and properscoring crps
        assert_allclose(actual, expected)
    # test that returns chunks
    assert_chunk(actual, chunk_bool)
    # test that attributes are kept
    assert_keep_attrs(actual, o, keep_attrs)
    # test that input types equal output types
    assign_type_input_output(actual, o)


@pytest.mark.slow
@pytest.mark.parametrize("keep_attrs", [True, False])
def test_crps_quadrature_args(o_dask, f_prob_dask, keep_attrs):
    """Test crps_quadrature args."""
    # to speed things up
    o_dask = o_dask.isel(time=0, drop=True)
    f_prob_dask = f_prob_dask.isel(time=0, drop=True)
    xmin, xmax, tol = -10, 10, 1e-6
    cdf_or_dist = norm
    actual = crps_quadrature(
        o_dask, cdf_or_dist, xmin, xmax, tol, keep_attrs=keep_attrs
    )
    expected = properscoring.crps_quadrature(o_dask, cdf_or_dist, xmin, xmax, tol)
    expected = xr.DataArray(expected, coords=o_dask.coords).mean()
    # test for numerical identity of xskillscore crps and crps
    assert_allclose(actual, expected)
    # test that xskillscore crps_ensemble returns chunks
    assert_chunk(actual, True)


@pytest.mark.slow
@pytest.mark.parametrize("dim", DIMS)
def test_crps_quadrature_dim(o, dim):
    """Check that crps_ensemble reduces only dim."""
    # to speed things up
    o = o.isel(time=0, drop=True)
    cdf_or_dist = norm
    actual = crps_quadrature(o, cdf_or_dist, dim=dim)
    assert_only_dim_reduced(dim, actual, o)


@pytest.mark.parametrize("input_type", ["float", "int", "numpy.array", "xr.DataArray"])
@pytest.mark.parametrize("keep_attrs", [True, False])
def test_crps_gaussian_args(o, keep_attrs, input_type):
    """Test that crps_gaussian accepts various data types as args."""
    mu = 0
    sig = 1
    if "input_type" == "float":
        mu = float(mu)
        sig = float(sig)
    elif "input_type" == "xr.DataArray":
        mu = xr.DataArray(mu)
        sig = xr.DataArray(sig)
    elif "input_type" == "np.array":
        mu = np.array(mu)
        sig = np.array(sig)
    actual = crps_gaussian(o, mu, sig, keep_attrs=keep_attrs)
    assert_keep_attrs(actual, o, keep_attrs)


@pytest.mark.parametrize("chunk_bool", [True, False])
@pytest.mark.parametrize("input_type", ["Dataset", "multidim Dataset", "DataArray"])
@pytest.mark.parametrize("keep_attrs", [True, False])
def test_threshold_brier_score_api_and_inputs(
    o, f_prob, keep_attrs, input_type, chunk_bool
):
    """Test that threshold_brier_score keeps attributes, chunking, input types and
    equals properscoring.threshold_brier_score."""
    o, f_prob = modify_inputs(o, f_prob, input_type, chunk_bool)
    threshold = 0.5
    actual = threshold_brier_score(o, f_prob, threshold, keep_attrs=keep_attrs)
    if input_type == "DataArray":  # properscoring allows only DataArrays
        expected = properscoring.threshold_brier_score(o, f_prob, threshold, axis=0)
        expected = xr.DataArray(expected, coords=o.coords).mean()
        expected["threshold"] = threshold
        # test for numerical identity of xs threshold and properscoring threshold
        if keep_attrs:
            expected = expected.assign_attrs(**actual.attrs)
        assert_identical(actual, expected)
    # test that returns chunks
    assert_chunk(actual, chunk_bool)
    # test that attributes are kept
    assert_keep_attrs(actual, o, keep_attrs)
    # test that input types equal output types
    assign_type_input_output(actual, o)


@pytest.mark.parametrize("dim", DIMS)
def test_threshold_brier_score_dim(o, f_prob, dim):
    """Check that threshold_brier_score reduces only dim."""
    actual = threshold_brier_score(o, f_prob, threshold=0.5, dim=dim)
    assert_only_dim_reduced(dim, actual, o)


@pytest.mark.parametrize(
    "threshold",
    [0, 0.5, [0.1, 0.3, 0.5], xr.DataArray([0.1, 0.3, 0.5], dims="threshold")],
    ids=["int", "flat", "list", "xr.DataArray"],
)
def test_threshold_brier_score_threshold(o_dask, f_prob_dask, threshold):
    """Test that threshold_brier_score accepts different kinds of thresholds."""
    actual = threshold_brier_score(o_dask, f_prob_dask, threshold)
    assert (actual.threshold == threshold).all()


@pytest.mark.skip(reason="not implemented")
def test_threshold_brier_score_threshold_dataset(o_dask, f_prob_dask):
    """Test that threshold_brier_score accepts xr.Dataset thresholds."""
    threshold = xr.DataArray([0.1, 0.3, 0.5], dims="threshold").to_dataset(name="var")
    threshold["var2"] = xr.DataArray([0.2, 0.4, 0.6], dims="threshold")
    o_dask = o_dask.to_dataset(name="var")
    o_dask["var2"] = o_dask["var"] ** 2
    f_prob_dask = f_prob_dask.to_dataset(name="var")
    f_prob_dask["var2"] = f_prob_dask["var"] ** 2
    actual = threshold_brier_score(o_dask, f_prob_dask, threshold)
    assert actual
    # test thresholds in each DataArray
    # assert (actual.threshold == threshold).all()


@pytest.mark.parametrize("chunk_bool", [True, False])
@pytest.mark.parametrize("input_type", ["Dataset", "multidim Dataset", "DataArray"])
@pytest.mark.parametrize("fair_bool", [True, False])
@pytest.mark.parametrize("keep_attrs", [True, False])
def test_brier_score_api_and_inputs(
    o, f_prob, keep_attrs, fair_bool, chunk_bool, input_type
):
    """Test that brier_score keeps attributes, chunking, input types."""
    o, f_prob = modify_inputs(o, f_prob, input_type, chunk_bool)
    f_prob > 0.5
    if not fair_bool:
        f_prob = f_prob.mean("member")
    o = (o > 0.5).assign_attrs(**o.attrs)
    actual = brier_score(
        o,
        f_prob,
        keep_attrs=keep_attrs,
        fair=fair_bool,
    )
    # test that returns chunks
    assert_chunk(actual, chunk_bool)
    # test that attributes are kept
    assert_keep_attrs(actual, o, keep_attrs)
    # test that input types equal output types
    assign_type_input_output(actual, o)


@pytest.mark.parametrize("fair_bool", [True, False])
@pytest.mark.parametrize("dim", DIMS)
def test_brier_score_dim(o, f_prob, dim, fair_bool):
    """Check that brier_score reduces only dim."""
    f_prob > 0.5
    if not fair_bool:
        f_prob = f_prob.mean("member")
    o = o > 0.5
    actual = brier_score(o, f_prob, dim=dim, fair=fair_bool)
    assert_only_dim_reduced(dim, actual, o)


def test_brier_score_forecast_member_dim(o, f_prob):
    """Check that brier_score allows forecasts with member dim and binary/boolean
    data."""
    f_prob > 0.5
    o = o > 0.5
    assert_identical(brier_score(o, f_prob), brier_score(o, f_prob.mean("member")))


@pytest.mark.parametrize("dim", DIMS)
def test_brier_score_vs_fair_brier_score(o, f_prob, dim):
    """Test that brier_score scores lower for limited ensemble than bias brier_score."""
    fbs = brier_score((o > 0.5), (f_prob > 0.5), dim=dim, fair=True)
    bs = brier_score((o > 0.5), (f_prob > 0.5).mean("member"), dim=dim, fair=False)
    assert (fbs <= bs).all(), print("fairBS", fbs, "\nBS", bs)


@pytest.mark.parametrize("chunk_bool", [True, False])
@pytest.mark.parametrize("input_type", ["DataArray", "Dataset", "multidim Dataset"])
@pytest.mark.parametrize("dim", DIMS)
def test_rank_histogram_sum(o, f_prob, dim, chunk_bool, input_type):
    """Test that the number of samples in the rank histogram is correct"""
    o, f_prob = modify_inputs(o, f_prob, input_type, chunk_bool)
    if dim == []:
        with pytest.raises(ValueError):
            rank_histogram(o, f_prob, dim=dim)
    else:
        rank_hist = rank_histogram(o, f_prob, dim=dim)
        if "Dataset" in input_type:
            rank_hist = rank_hist[list(o.data_vars)[0]]
            o = o[list(o.data_vars)[0]]
            assert_allclose(rank_hist.sum(), o.count())
        assert_allclose(rank_hist.sum(), o.count())
        # test that returns chunks
        assert_chunk(rank_hist, chunk_bool)
        # test that attributes are kept # TODO: add
        # assert_keep_attrs(rank_hist, o, keep_attrs)
        # test that input types equal output types
        assign_type_input_output(rank_hist, o)


def test_rank_histogram_values(o, f_prob):
    """Test values in extreme cases that observations \
        all smaller/larger than forecasts"""
    assert rank_histogram((f_prob.min() - 1) + 0 * o, f_prob)[0] == o.size
    assert rank_histogram((f_prob.max() + 1) + 0 * o, f_prob)[-1] == o.size


@pytest.mark.parametrize("dim", DIMS)
@pytest.mark.parametrize("chunk_bool", [True, False])
@pytest.mark.parametrize("input_type", ["DataArray", "Dataset", "multidim Dataset"])
def test_discrimination_sum(o, f_prob, dim, chunk_bool, input_type):
    """Test that the probabilities sum to 1"""
    o, f_prob = modify_inputs(o, f_prob, input_type, chunk_bool)
    if dim == []:
        with pytest.raises(ValueError):
            discrimination(o > 0.5, (f_prob > 0.5).mean("member"), dim=dim)
    else:
        disc = discrimination(o > 0.5, (f_prob > 0.5).mean("member"), dim=dim)
        # test that input types equal output types
        assign_type_input_output(disc, o)
        if "Dataset" in input_type:
            disc = disc[list(o.data_vars)[0]]
        # dont understand the error message here, but it appeared
        with suppress_warnings("invalid value encountered in true_divide"):
            hist_event_sum = (
                disc.sel(event=True).sum("forecast_probability", skipna=False).values
            )
            hist_no_event_sum = (
                disc.sel(event=False).sum("forecast_probability", skipna=False).values
            )
        # Note, xarray's assert_allclose is already imported but won't compare to scalar
        assert np.allclose(hist_event_sum[~np.isnan(hist_event_sum)], 1)
        assert np.allclose(hist_no_event_sum[~np.isnan(hist_no_event_sum)], 1)

        # test that returns chunks
        assert_chunk(disc, chunk_bool)
        # test that attributes are kept # TODO: add
        # assert_keep_attrs(disc, o, keep_attrs)


def test_discrimination_perfect_values(o):
    """Test values for perfect forecast"""
    f = xr.concat(10 * [o], dim="member")
    disc = discrimination(o > 0.5, (f > 0.5).mean("member"))
    assert np.allclose(disc.sel(event=True)[-1], 1)
    assert np.allclose(disc.sel(event=True)[:-1], 0)
    assert np.allclose(disc.sel(event=False)[0], 1)
    assert np.allclose(disc.sel(event=False)[1:], 0)


@pytest.mark.parametrize("dim", DIMS)
@pytest.mark.parametrize("chunk_bool", [True, False])
@pytest.mark.parametrize("input_type", ["DataArray", "Dataset", "multidim Dataset"])
def test_reliability_api_and_inputs(o, f_prob, dim, chunk_bool, input_type):
    """Test that reliability keeps chunking and input types."""
    o, f_prob = modify_inputs(o, f_prob, input_type, chunk_bool)
    if dim == []:
        with pytest.raises(ValueError):
            reliability(o > 0.5, (f_prob > 0.5).mean("member"), dim)
    else:
        actual = reliability(o > 0.5, (f_prob > 0.5).mean("member"), dim=dim)
        # test that returns chunks
        assert_chunk(actual, chunk_bool)
        # test that attributes are kept
        # assert_keep_attrs(actual, o, keep_attrs) # TODO: implement
        # test that input types equal output types
        assign_type_input_output(actual, o)


def test_reliability_values(o, f_prob):
    """Test 1D reliability values against sklearn calibration_curve"""
    for lon in f_prob.lon:
        for lat in f_prob.lat:
            o_1d = o.sel(lon=lon, lat=lat) > 0.5
            f_1d = (f_prob.sel(lon=lon, lat=lat) > 0.5).mean("member")
            # scipy bins are only left-edge inclusive and 1e-8 is added to the last bin, whereas
            # xhistogram the rightmost edge of xhistogram bins is included - mimic scipy behaviour
            actual = reliability(
                o_1d, f_1d, probability_bin_edges=np.linspace(0, 1 + 1e-8, 6)
            )
            expected, _ = calibration_curve(
                o_1d, f_1d, normalize=False, n_bins=5, strategy="uniform"
            )
            npt.assert_allclose(actual.where(actual.notnull(), drop=True), expected)
            npt.assert_allclose(actual["samples"].sum(), o_1d.size)


def test_reliability_perfect_values(o):
    """Test values for perfect forecast"""
    f_prob = xr.concat(10 * [o], dim="member")
    actual = reliability(
        o > 0.5,
        (f_prob > 0.5).mean("member"),
    )
    expected_true_samples = (o > 0.5).sum()
    expected_false_samples = (o <= 0.5).sum()
    assert np.allclose(actual[0], 0)
    assert np.allclose(actual[-1], 1)
    assert np.allclose(actual["samples"][0], expected_false_samples)
    assert np.allclose(actual["samples"][-1], expected_true_samples)
    assert np.allclose(actual["samples"].sum(), o.size)


@pytest.mark.parametrize("fair_bool", [True, False])
@pytest.mark.parametrize("dim", DIMS)
def test_rps_reduce_dim(o, f_prob, category_edges, dim, fair_bool):
    """Test that rps reduced dim and works for (chunked) ds and da"""
    actual = rps(o, f_prob, category_edges=category_edges, dim=dim, fair=fair_bool)
    assert_only_dim_reduced(dim, actual, o)


def test_rps_category_edges_None_fails(o, f_prob):
    """Test that rps expects inputs to have category_edges dim if category_edges is None."""
    with pytest.raises(ValueError, match="Expected dimension"):
        rps(o, f_prob, category_edges=None, dim=[], input_distributions="c")


@pytest.mark.parametrize("input_distributions", ["c", "p"])
def test_rps_category_edges_None_works(o, f_prob, input_distributions):
    """Test that rps expects inputs to have category_edges dim if category_edges is None."""
    o = o.rename({"time": "category"})
    f_prob = f_prob.rename({"time": "category"}).mean("member")
    rps(o, f_prob, category_edges=None, dim=[], input_distributions=input_distributions)


@pytest.mark.parametrize("chunk_bool", [True, False])
@pytest.mark.parametrize("input_type", ["Dataset", "multidim Dataset", "DataArray"])
@pytest.mark.parametrize("keep_attrs", [True, False])
def test_rps_api_and_inputs(
    o, f_prob, category_edges, keep_attrs, input_type, chunk_bool
):
    """Test that rps keeps attributes, chunking, input types."""
    o, f_prob = modify_inputs(o, f_prob, input_type, chunk_bool)
    category_edges = xr.DataArray(category_edges, dims="category_edge")
    if "Dataset" in input_type:
        category_edges = category_edges.to_dataset(name="var")
        if "multidim" in input_type:
            category_edges["var2"] = category_edges["var"]
    actual = rps(o, f_prob, category_edges, keep_attrs=keep_attrs)
    # test that returns chunks
    assert_chunk(actual, chunk_bool)
    # test that attributes are kept
    assert_keep_attrs(actual, o, keep_attrs)
    # test that input types equal output types
    assign_type_input_output(actual, o)


def rps_xhist(
    observations,
    forecasts,
    category_edges,
    dim=None,
    fair=False,
    weights=None,
    keep_attrs=False,
    member_dim="member",
):
    """Old way to calculate RPS with xhistogram.

    category_edges : array_like, xr.Dataset, xr.DataArray, None

        - array_like: Category bin edges used to compute the CDFs. Similar to
          np.histogram, all but the last (righthand-most) bin include the
          left edge and exclude the right edge. The last bin includes both edges.
          CDFs based on boolean or logical (True or 1 for event occurance, False or 0
          for non-occurance) observations.
          If ``fair==False``, forecasts should be between 0 and 1 without a dimension
          ``member_dim`` or boolean / binary containing a member dimension
          (probabilities will be internally calculated by ``.mean(member_dim))``.
          If ``fair==True``, forecasts must be boolean / binary containing dimension
          ``member_dim``."""
    from xskillscore.core.contingency import _get_category_bounds
    from xskillscore.core.utils import _keep_nans_masked, histogram

    bin_names = ["category"]
    bin_dim = f"{bin_names[0]}_edge"
    M = forecasts[member_dim].size

    assert isinstance(category_edges, np.ndarray)

    # histogram(dim=[]) not allowed therefore add fake member dim
    # to apply over when multi-dim observations
    if len(observations.dims) == 1:
        observations_bins = histogram(
            observations,
            bins=[category_edges],
            bin_names=["category_edge"],
            dim=None,
        )
    else:
        observations_bins = histogram(
            observations.expand_dims(member_dim),
            bins=[category_edges],
            bin_names=["category_edge"],
            dim=[member_dim],
        )
    if "category_edge_bin" in observations_bins.dims:
        observations_bins = observations_bins.rename(
            {"category_edge_bin": "category_edge"}
        )

    forecasts = histogram(
        forecasts,
        bins=[category_edges],
        bin_names=["category_edge"],
        dim=[member_dim],
    )
    if "category_edge_bin" in forecasts.dims:
        forecasts = forecasts.rename({"category_edge_bin": "category_edge"})

    # normalize f.sum()=1 to make cdf
    forecasts = forecasts / forecasts.sum(bin_dim)

    Fc = forecasts.cumsum(bin_dim)
    Oc = observations_bins.cumsum(bin_dim)

    # RPS formulas
    if fair:
        Ec = Fc * M
        res = ((Ec / M - Oc) ** 2 - Ec * (M - Ec) / (M ** 2 * (M - 1))).sum(bin_dim)
    else:
        res = ((Fc - Oc) ** 2).sum(bin_dim)

    if weights is not None:
        res = res.weighted(weights)

    res = res.mean(dim, keep_attrs=keep_attrs)
    # add bin edges as coords
    res = res.assign_coords(
        {"forecasts_category_edge": ", ".join(_get_category_bounds(category_edges))}
    )
    res = res.assign_coords(
        {"observations_category_edge": ", ".join(_get_category_bounds(category_edges))}
    )

    # keep nans and prevent 0 for all nan grids
    res = _keep_nans_masked(observations, res, dim, ignore=["category_edge"])
    return res


def test_rps_wilks_example():
    """Test with values from Wilks, D. S. (2006). Statistical methods in the
    atmospheric sciences (2nd ed, Vol. 91). Amsterdam ; Boston: Academic Press. p.301.
    """
    category_edges = np.array([0.0, 0.01, 0.24, 1.0])
    # first example
    # xhistogram way with np.array category_edges
    Obs = xr.DataArray([0.0001])  # .expand_dims('time')  # no precip
    F1 = xr.DataArray(
        [0] * 2 + [0.1] * 5 + [0.3] * 3, dims="member"
    )  # .expand_dims('time')
    F2 = xr.DataArray(
        [0] * 2 + [0.1] * 3 + [0.3] * 5, dims="member"
    )  # .expand_dims('time')
    np.testing.assert_allclose(rps_xhist(Obs, F1, category_edges), 0.73)
    np.testing.assert_allclose(rps_xhist(Obs, F2, category_edges), 0.89)
    # xr way with xr.DataArray category_edges
    xr_category_edges = xr.DataArray(
        category_edges, dims="category_edge", coords={"category_edge": category_edges}
    )
    assert_allclose(rps(Obs, F1, category_edges), rps(Obs, F1, xr_category_edges))
    assert_allclose(rps(Obs, F2, category_edges), rps(Obs, F2, xr_category_edges))

    # second example
    Obs = xr.DataArray([0.3])  # larger than 0.25
    np.testing.assert_allclose(rps_xhist(Obs, F1, category_edges), 0.53)
    np.testing.assert_allclose(rps_xhist(Obs, F2, category_edges), 0.29)
    # xr way with xr.DataArray category_edges
    assert_allclose(rps(Obs, F1, category_edges), rps(Obs, F1, xr_category_edges))
    assert_allclose(rps(Obs, F2, category_edges), rps(Obs, F2, xr_category_edges))


def test_rps_wilks_example_pdf():
    """Test xs.rps(category_edges=None, input_distributions='p') with values from
    Wilks, D. S. (2006). Statistical methods in the atmospheric sciences (2nd ed,
    Vol. 91). Amsterdam ; Boston: Academic Press. p.301.
    """
    Obs = xr.DataArray([1.0, 0.0, 0.0], dims="category")  # no precip
    F1 = xr.DataArray([0.2, 0.5, 0.3], dims="category")
    F2 = xr.DataArray([0.2, 0.3, 0.5], dims="category")
    np.testing.assert_allclose(
        rps(Obs, F1, category_edges=None, input_distributions="p"), 0.73
    )
    np.testing.assert_allclose(
        rps(Obs, F2, category_edges=None, input_distributions="p"), 0.89
    )

    # second example
    Obs = xr.DataArray([0.0, 0.0, 1.0], dims="category")  # larger than 0.25
    np.testing.assert_allclose(
        rps(Obs, F1, category_edges=None, input_distributions="p"), 0.53
    )
    np.testing.assert_allclose(
        rps(Obs, F2, category_edges=None, input_distributions="p"), 0.29
    )


@pytest.mark.parametrize("fair_bool", [True, False])
def test_2_category_rps_equals_brier_score(o, f_prob, fair_bool):
    """Test that RPS for two categories equals the Brier Score."""
    category_edges = np.array([0.0, 0.5, 1.0])
    assert_allclose(
        rps(
            o.rename({"time": "category"}),
            f_prob.rename({"time": "category"}),
            category_edges=category_edges,
            dim=None,
            fair=fair_bool,
        ).drop(["forecasts_category_edge", "observations_category_edge"]),
        brier_score(o > 0.5, (f_prob > 0.5), dim=None, fair=fair_bool),
    )


def test_rps_fair_category_edges_None(o, f_prob):
    """Test that RPS without category_edges works for fair==True if forecast[member]
    set."""
    rps(
        o.rename({"time": "category"}),
        f_prob.mean("member")
        .rename({"time": "category"})
        .assign_coords(member=f_prob.member.size),
        category_edges=None,
        dim=None,
        fair=True,
        input_distributions="p",
    )


@pytest.mark.parametrize("fair_bool", [True, False])
def test_rps_perfect_values(o, category_edges, fair_bool):
    """Test values for perfect forecast"""
    f = xr.concat(10 * [o], dim="member")
    res = rps(o, f, category_edges=category_edges, fair=fair_bool)
    assert (res == 0).all()


@pytest.mark.parametrize("fair_bool", [True, False])
def test_rps_dask(o_dask, f_prob_dask, category_edges, fair_bool):
    """Test that rps returns dask array if provided dask array"""
    assert (
        rps(o_dask, f_prob_dask, category_edges=category_edges, fair=fair_bool).chunks
        is not None
    )


@pytest.mark.parametrize("dim", DIMS)
def test_rps_vs_fair_rps(o, f_prob, category_edges, dim):
    """Test that fair rps is smaller (e.g. better) or equal than rps due to ensemble-
    size adjustment."""
    frps = rps(o, f_prob, dim=dim, fair=True, category_edges=category_edges)
    ufrps = rps(o, f_prob, dim=dim, fair=False, category_edges=category_edges)
    assert (frps <= ufrps).all()


@pytest.mark.parametrize("fair_bool", [True, False])
def test_rps_category_edges_xrDataArray(o, f_prob, fair_bool):
    """Test rps with category_edges as xrDataArray for forecast and observations edges."""
    category_edges = f_prob.quantile(
        q=[0.2, 0.4, 0.6, 0.8], dim=["time", "member"]
    ).rename({"quantile": "category_edge"})
    actual = rps(
        o,
        f_prob,
        dim="time",
        fair=fair_bool,
        category_edges=category_edges,
    )
    assert set(["lon", "lat"]) == set(actual.dims)
    assert "category_edge" not in actual.dims


@pytest.mark.parametrize("fair_bool", [True, False])
def test_rps_category_edges_xrDataset(o, f_prob, fair_bool):
    """Test rps with category_edges as xrDataArray for forecast and observations edges."""
    o = o.to_dataset(name="var")
    o["var2"] = o["var"] ** 2
    f_prob = f_prob.to_dataset(name="var")
    f_prob["var2"] = f_prob["var"] ** 2
    category_edges = f_prob.quantile(
        q=[0.2, 0.4, 0.6, 0.8], dim=["time", "member"]
    ).rename({"quantile": "category_edge"})
    actual = rps(
        o,
        f_prob,
        dim="time",
        fair=fair_bool,
        category_edges=category_edges,
    )
    assert set(["lon", "lat"]) == set(actual.dims)
    assert "category_edge" not in actual.dims


@pytest.mark.parametrize("fair_bool", [True, False])
def test_rps_category_edges_tuple(o, f_prob, fair_bool):
    """Test rps with category_edges as tuple of xrDataArray for forecast and observations edges separately."""
    o_edges = o.quantile(q=[0.2, 0.4, 0.6, 0.8], dim="time").rename(
        {"quantile": "category_edge"}
    )
    f_edges = f_prob.quantile(q=[0.2, 0.4, 0.6, 0.8], dim=["time", "member"]).rename(
        {"quantile": "category_edge"}
    )
    actual = rps(
        o,
        f_prob,
        dim="time",
        fair=fair_bool,
        category_edges=(o_edges, f_edges),
    )
    assert set(["lon", "lat"]) == set(actual.dims)
    assert "category_edge" not in actual.dims


@pytest.mark.parametrize("input_distributions", ["c", "p"])
def test_rps_category_edges_None(o, f_prob, input_distributions):
    """Test rps with category_edges as None expecting o and f_prob are already CDFs."""
    e = [0, 0.2, 0.4, 0.6, 0.8, 1.0]
    bin_dim = "category"
    edges = xr.DataArray(e, dims=bin_dim, coords={bin_dim: e})
    o_c = o < edges  # CDF
    f_c = (f_prob < edges).mean("member")  # CDF
    actual = rps(
        o_c,
        f_c,
        dim="time",
        fair=False,
        category_edges=None,
        input_distributions=input_distributions,
    )
    assert set(["lon", "lat"]) == set(actual.dims)
    assert "quantile" not in actual.dims
    assert "member" not in actual.dims


@pytest.mark.parametrize(
    "category_edges",
    [
        xr.DataArray(
            [0.2, 0.4, 0.6, 0.8],
            dims="category_edge",
            coords={"category_edge": [0.2, 0.4, 0.6, 0.8]},
        ),
        np.array([0.2, 0.4, 0.6, 0.8]),
    ],
    ids=["edge xr", "edge np"],
)
@pytest.mark.parametrize("fair_bool", [True, False], ids=["fair=True", "fair=False"])
def test_rps_keeps_masked(o, f_prob, fair_bool, category_edges):
    """Test rps keeps NaNs."""
    o = o.where(o.lat > 1)
    f_prob = f_prob.where(f_prob.lat > 1)
    actual = rps(o, f_prob, dim="time", category_edges=category_edges)
    assert set(["lon", "lat"]) == set(actual.dims)
    assert actual.isel(lat=[0, 1]).isnull().all()
    assert actual.isel(lat=slice(2, None)).notnull().all()
    # test forecasts_category_edge no repeats
    assert (
        "[-np.inf, 0.2), [0.2, 0.4), [0.4, 0.6), [0.6, 0.8), [0.8, np.inf]"
        in actual.coords["forecasts_category_edge"].values
    )
    # one more category internally used than category_edges provided
    assert len(category_edges) + 1 == str(
        actual.coords["forecasts_category_edge"].values
    ).count("[")


@pytest.mark.parametrize("fair_bool", [True, False], ids=["bool=fair", "fair=False"])
def test_rps_new_identical_old_xhistogram(o, f_prob, fair_bool):
    """Test that new rps algorithm is identical to old algorithm with xhistogram.
    Makes a difference whether full range of f_prob is covered or not."""
    category_edges_np = np.array([0, 0.2, 0.4, 0.6, 0.8, 1.0])
    category_edges_xr = xr.DataArray(
        category_edges_np,
        dims="category_edge",
        coords={"category_edge": category_edges_np},
    )
    dim = "time"
    actual = rps(o, f_prob, dim=dim, category_edges=category_edges_xr)
    expected = rps_xhist(o, f_prob, dim=dim, category_edges=category_edges_np)
    drop = ["observations_category_edge", "forecasts_category_edge"]
    assert_allclose(
        actual.rename("histogram_category_edge").drop(drop), expected.drop(drop)
    )


def test_rps_last_edge_included(o, f_prob):
    """Test that last edges is included."""
    category_edges_np = np.array([0, 0.2, 0.4, 0.6, 0.8, 1.0])
    o = xr.ones_like(o)
    f_prob = xr.ones_like(f_prob)
    res_actual = rps(o, f_prob, dim="time", category_edges=category_edges_np)
    assert (res_actual == 0).all()


@pytest.mark.parametrize(
    "observation,forecast",
    [
        (
            lazy_fixture("observation_1d_long"),
            lazy_fixture("forecast_1d_long"),
        ),
        (
            lazy_fixture("observation_3d"),
            lazy_fixture("forecast_3d"),
        ),
    ],
)
@pytest.mark.parametrize("dim", ["time", None])
@pytest.mark.parametrize("drop_intermediate_bool", [True, False])
@pytest.mark.parametrize("chunk", [True, False])
@pytest.mark.parametrize("input", ["Dataset", "DataArray"])
@pytest.mark.parametrize(
    "return_results", ["all_as_tuple", "area", "all_as_metric_dim"]
)
def test_roc_returns(
    observation,
    forecast,
    symmetric_edges,
    dim,
    return_results,
    input,
    chunk,
    drop_intermediate_bool,
):
    """testing keywords and inputs pass"""
    if "Dataset" in input:
        name = "var"
        forecast = forecast.to_dataset(name=name)
        observation = observation.to_dataset(name=name)
    if chunk:
        forecast = forecast.chunk()
        observation = observation.chunk()

    roc(
        observation,
        forecast,
        symmetric_edges,
        dim=dim,
        drop_intermediate=drop_intermediate_bool,
        return_results=return_results,
    )


def test_roc_auc_score_random_forecast(
    forecast_1d_long, observation_1d_long, symmetric_edges
):
    """Test that ROC AUC around 0.5 for random forecast."""
    area = roc(
        observation_1d_long,
        forecast_1d_long,
        symmetric_edges,
        dim="time",
        return_results="area",
    )
    assert area < 0.65
    assert area > 0.35


def test_roc_auc_score_perfect_forecast(forecast_1d_long, symmetric_edges):
    """Test that ROC AUC equals 1 for perfect forecast."""
    area = roc(
        forecast_1d_long,
        forecast_1d_long,
        symmetric_edges,
        drop_intermediate=False,
        dim="time",
        return_results="area",
    )
    assert area == 1.0


@pytest.mark.parametrize("drop_intermediate_bool", [False, True])
def test_roc_auc_score_out_of_range_forecast(
    forecast_1d_long, observation_1d_long, symmetric_edges, drop_intermediate_bool
):
    """Test that ROC AUC equals 0.0 for out of range forecast."""
    area = roc(
        observation_1d_long,
        xr.ones_like(forecast_1d_long) + 100,
        symmetric_edges,
        drop_intermediate=drop_intermediate_bool,
        dim="time",
        return_results="area",
    )
    assert float(area) in [0.0, 0.5]  # expect 0.0 but sometimes evaluates at 0.5 in CI


@pytest.mark.parametrize("drop_intermediate_bool", [False, True])
def test_roc_auc_score_out_of_range_observation(
    forecast_1d_long, observation_1d_long, symmetric_edges, drop_intermediate_bool
):
    """Test that ROC AUC equals 0.0 for out of range observation."""
    area = roc(
        xr.ones_like(observation_1d_long) + 100,
        forecast_1d_long,
        symmetric_edges,
        drop_intermediate=drop_intermediate_bool,
        dim="time",
        return_results="area",
    )
    np.testing.assert_almost_equal(area, 0.0, decimal=2)


@pytest.mark.parametrize("drop_intermediate_bool", [False, True])
def test_roc_auc_score_out_of_range_edges(
    forecast_1d_long, observation_1d_long, symmetric_edges, drop_intermediate_bool
):
    """Test that ROC AUC equals 0.5 for out of range edges."""
    area = roc(
        observation_1d_long,
        forecast_1d_long,
        symmetric_edges + 100,
        drop_intermediate=drop_intermediate_bool,
        dim="time",
        return_results="area",
    )
    assert float(area) == 0.5


@pytest.mark.parametrize("drop_intermediate_bool", [False, True])
def test_roc_auc_score_constant_forecast(
    forecast_1d_long, observation_1d_long, symmetric_edges, drop_intermediate_bool
):
    """Test that ROC AUC equals 0.5 for constant forecast."""
    xs_area = roc(
        observation_1d_long > 0,
        forecast_1d_long * 0,
        symmetric_edges,
        drop_intermediate=drop_intermediate_bool,
        dim="time",
        return_results="area",
    )
    sk_area = roc_auc_score(observation_1d_long > 0, forecast_1d_long * 0)
    np.testing.assert_allclose(xs_area, sk_area)


@pytest.mark.parametrize("drop_intermediate_bool", [False, True])
def test_roc_bin_edges_continuous_against_sklearn(
    forecast_1d_long, observation_1d_long, drop_intermediate_bool
):
    """Test xs.roc against sklearn.metrics.roc_curve/auc_score."""
    fp = forecast_1d_long.clip(0, 1)  # prob
    ob = observation_1d_long > 0  # binary
    # sklearn
    sk_fpr, sk_tpr, _ = roc_curve(ob, fp, drop_intermediate=drop_intermediate_bool)
    sk_area = roc_auc_score(ob, fp)
    # xs
    xs_fpr, xs_tpr, xs_area = roc(
        ob,
        fp,
        "continuous",
        drop_intermediate=drop_intermediate_bool,
        return_results="all_as_tuple",
    )
    np.testing.assert_allclose(xs_area, sk_area)
    if not drop_intermediate_bool:  # drops sometimes one too much or too little
        assert (xs_fpr == sk_fpr).all()
        assert (xs_tpr == sk_tpr).all()


def test_roc_bin_edges_drop_intermediate(forecast_1d_long, observation_1d_long):
    """Test that drop_intermediate reduces probability_bins in xs.roc ."""
    fp = forecast_1d_long.clip(0, 1)  # prob
    ob = observation_1d_long > 0  # binary
    # xs
    txs_fpr, txs_tpr, txs_area = roc(
        ob, fp, "continuous", drop_intermediate=True, return_results="all_as_tuple"
    )
    fxs_fpr, fxs_tpr, fxs_area = roc(
        ob, fp, "continuous", drop_intermediate=False, return_results="all_as_tuple"
    )
    # same area
    np.testing.assert_allclose(fxs_area, txs_area)
    # same or less probability_bins
    assert len(fxs_fpr) >= len(txs_fpr)
    assert len(fxs_tpr) >= len(txs_tpr)


def test_roc_keeps_probability_bin_as_coord(
    observation_1d_long, forecast_1d_long, symmetric_edges
):
    """Test that roc keeps probability_bin as coords."""
    fpr, tpr, area = roc(
        observation_1d_long,
        forecast_1d_long,
        symmetric_edges,
        drop_intermediate=False,
        return_results="all_as_tuple",
    )
    assert (tpr.probability_bin == symmetric_edges).all()
    assert (fpr.probability_bin == symmetric_edges).all()


def test_roc_bin_edges_symmetric_asc_or_desc(
    observation_1d_long, forecast_1d_long, symmetric_edges
):
    """Test that roc bin_edges works increasing or decreasing order."""
    fpr, tpr, area = roc(
        observation_1d_long,
        forecast_1d_long,
        symmetric_edges,
        drop_intermediate=False,
        return_results="all_as_tuple",
    )
    fpr2, tpr2, area2 = roc(
        observation_1d_long,
        forecast_1d_long,
        symmetric_edges[::-1],
        drop_intermediate=False,
        return_results="all_as_tuple",
    )
    assert_identical(fpr, fpr2.sortby(fpr.probability_bin))
    assert_identical(tpr, tpr2.sortby(tpr.probability_bin))
    assert_identical(area, area2)



import dask
import numpy as np
import xarray as xr

CONCAT_KWARGS = {"coords": "minimal", "compat": "override"}


def _gen_idx(forecast, dim, iterations, select_dim_items, replace, new_dim):
    """Generate indices to select from. Replace decides whether resampling is with or
    without replacement.

    Args: (selected)
        select_dim_items(int) : number of new items in dim to select
        new_dim (xr.DataArray) : new dimension of above
    """
    if replace:
        idx = np.random.randint(0, forecast[dim].size, (iterations, select_dim_items))
    elif not replace:
        # create 2d np.arange
        idx = np.linspace(
            (np.arange(select_dim_items)),
            (np.arange(select_dim_items)),
            iterations,
            dtype="int",
        )
        # shuffle each line
        for ndx in np.arange(iterations):
            np.random.shuffle(idx[ndx])
    idx_da = xr.DataArray(
        idx,
        dims=("iteration", dim),
        coords=({"iteration": range(iterations), dim: new_dim}),
    )
    return idx_da


def resample_iterations(forecast, iterations, dim="member", dim_max=None, replace=True):
    """Resample over ``dim`` by index ``iterations`` times.

    .. note::
        This gives the same result as
        :py:func:`~xskillscore.resampling.resample_iterations_idx` but slower.
        When using ``dask``, the number of tasks in
        :py:func:`~xskillscore.resampling.resample_iterations` will scale with
        ``iterations`` but constant chunksize, whereas the tasks in
        :py:func:`~xskillscore.resampling.resample_iterations_idx` will stay constant
        with increasing chunksize.

    Parameters
    ----------
    forecast : xr.DataArray, xr.Dataset
        Forecast.
    iterations : int
        Number of resampling iterations.
    dim : str
        Dimension name to resample over. Defaults to ``'member'``.
    dim_max : int
        Number of items to select in `dim`.
    replace : bool
        Resampling with or without replacement. Defaults to ``True``.

    Returns
    -------
    forecast_smp : xr.DataArray, xr.Dataset
        data resampled along dimension ``dim`` with additional ``dim='iteration'``.

    Examples
    --------
    >>> a = xr.DataArray(np.random.rand(1000, 3, 3), dims=['time', 'x', 'y'])
    >>> xs.resample_iterations(a, 500, 'time')  # doctest: +SKIP
    <xarray.DataArray (time: 1000, x: 3, y: 3, iteration: 500)>

    See also
    --------
    :py:func:`~xskillscore.resampling.resample_iterations_idx`

    References
    ----------
    * Mason, S. J., & Mimmack, G. M. (1992). The use of bootstrap confidence intervals
      for the correlation coefficient in climatology. Theoretical and Applied
      Climatology, 45(4), 229–233. https://doi.org/10/b6fnsv

    * Mason, S. J. (2008). Understanding forecast verification statistics.
      Meteorological Applications, 15(1), 31–40. https://doi.org/10/bgvgnz

    * Goddard, L., Kumar, A., Solomon, A., Smith, D., Boer, G., Gonzalez, P.,
      Kharin, V., Merryfield, W., Deser, C., Mason, S. J., Kirtman, B. P., Msadek, R.,
      Sutton, R., Hawkins, E., Fricker, T., Hegerl, G., Ferro, C. a. T.,
      Stephenson, D. B., Meehl, G. A., … Delworth, T. (2013). A verification framework
      for interannual-to-decadal predictions experiments. Climate Dynamics, 40(1–2),
      245–272. https://doi.org/10/f4jjvf
    """
    if dim_max is not None and dim_max <= forecast[dim].size:
        # select only dim_max items
        select_dim_items = dim_max
        new_dim = forecast[dim].isel({dim: slice(None, dim_max)})
    else:
        select_dim_items = forecast[dim].size
        new_dim = forecast[dim]

    # generate random indices to select from
    idx_da = _gen_idx(forecast, dim, iterations, select_dim_items, replace, new_dim)
    # select those indices in a loop
    forecast_smp = []
    for i in np.arange(iterations):
        idx = idx_da.sel(iteration=i).data
        forecast_smp.append(forecast.isel({dim: idx}).assign_coords({dim: new_dim}))
    forecast_smp = xr.concat(forecast_smp, dim="iteration", **CONCAT_KWARGS)
    forecast_smp["iteration"] = np.arange(iterations)
    return forecast_smp.transpose(..., "iteration")


def resample_iterations_idx(
    forecast, iterations, dim="member", replace=True, dim_max=None
):
    """Resample over ``dim`` by index ``iterations`` times.

    .. note::
        This is a much faster way to bootstrap/resample each iteration
        individually and applying the function to it. This will create a
        DataArray with dimension ``iteration`` of size ``iterations``.
        When using ``dask``, the number of tasks in
        :py:func:`~xskillscore.resampling.resample_iterations` will scale with
        ``iterations`` but constant chunksize, whereas the tasks in
        :py:func:`~xskillscore.resampling.resample_iterations_idx` will stay constant
        with increasing chunksize.

    Parameters
    ----------
        forecast : xr.DataArray, xr.Dataset
            Forecast.
        iterations : int
            Number of resampling iterations.
        dim : str
            Dimension name to resample over. Defaults to ``'member'``.
        replace : bool
            Resampling with or without replacement. Defaults to ``True``.
        dim_max : int
            Number of item from ``dim`` to return.

    Returns
    -------
    forecast_smp : xr.DataArray, xr.Dataset
        data resampled along dimension ``dim`` with additional ``dim='iteration'``.

    Examples
    --------
    >>> a = xr.DataArray(np.random.rand(1000, 3, 3),
    ...                  coords=[("time", np.arange(1000)),
    ...                          ("x", np.arange(3)),
    ...                          ("y", np.arange(3))])
    >>> xs.resample_iterations_idx(a, 500, 'time') # doctest: +SKIP
    <xarray.DataArray (time: 1000, x: 3, y: 3, iteration: 500)>

    See also
    --------
    :py:func:`~xskillscore.resampling.resample_iterations`

    References
    ----------
    * Mason, S. J., & Mimmack, G. M. (1992). The use of bootstrap confidence intervals
      for the correlation coefficient in climatology. Theoretical and Applied
      Climatology, 45(4), 229–233. https://doi.org/10/b6fnsv

    * Mason, S. J. (2008). Understanding forecast verification statistics.
      Meteorological Applications, 15(1), 31–40. https://doi.org/10/bgvgnz

    * Goddard, L., Kumar, A., Solomon, A., Smith, D., Boer, G., Gonzalez, P.,
      Kharin, V., Merryfield, W., Deser, C., Mason, S. J., Kirtman, B. P., Msadek, R.,
      Sutton, R., Hawkins, E., Fricker, T., Hegerl, G., Ferro, C. a. T.,
      Stephenson, D. B., Meehl, G. A., … Delworth, T. (2013). A verification framework
      for interannual-to-decadal predictions experiments. Climate Dynamics, 40(1–2),
      245–272. https://doi.org/10/f4jjvf
    """
    # equivalent to above
    select_dim_items = forecast[dim].size
    new_dim = forecast[dim]

    def select_bootstrap_indices_ufunc(x, idx):
        """Selects multi-level indices ``idx`` from xarray object ``x`` for all
        iterations."""
        # `apply_ufunc` sometimes adds a singleton dimension on the end, so we squeeze
        # it out here. This leverages multi-level indexing from numpy, so we can
        # select a different set of, e.g., ensemble members for each iteration and
        # construct one large DataArray with ``iterations`` as a dimension.
        return np.moveaxis(x.squeeze()[idx.squeeze().transpose()], 0, -1)

    # generate random indices to select from
    idx_da = _gen_idx(forecast, dim, iterations, select_dim_items, replace, new_dim)
    # select those indices in one go
    transpose_kwargs = (
        {"transpose_coords": False} if isinstance(forecast, xr.DataArray) else {}
    )
    # bug fix when only one iteration
    if iterations == 1:
        forecast_smp = forecast.isel({dim: idx_da.isel(iteration=0, drop=True).values})
    else:
        forecast_smp = xr.apply_ufunc(
            select_bootstrap_indices_ufunc,
            forecast.transpose(dim, ..., **transpose_kwargs),
            idx_da,
            dask="parallelized",
            output_dtypes=[float],
        )
    # return only dim_max members
    if dim_max is not None and dim_max <= forecast[dim].size:
        forecast_smp = forecast_smp.isel({dim: slice(None, dim_max)})
    return forecast_smp



import numpy as np
import pytest
import xarray as xr

import xskillscore as xs
from xskillscore import Contingency

PERIODS = 12  # effective_p_value produces nans for shorter periods


@pytest.fixture(autouse=True)
def add_standard_imports(doctest_namespace):
    """imports for doctest"""
    doctest_namespace["np"] = np
    doctest_namespace["xr"] = xr
    doctest_namespace["xs"] = xs

    # always seed numpy.random to make the examples deterministic
    np.random.seed(42)


@pytest.fixture
def times():
    return xr.cftime_range(start="2000", periods=PERIODS, freq="D")


@pytest.fixture
def lats():
    return np.arange(4)


@pytest.fixture
def lons():
    return np.arange(5)


@pytest.fixture
def members():
    return np.arange(3)


# o vs. f in probabilistic
@pytest.fixture
def o(times, lats, lons):
    """Observation."""
    data = np.random.rand(len(times), len(lats), len(lons))
    return xr.DataArray(
        data,
        coords=[times, lats, lons],
        dims=["time", "lat", "lon"],
        attrs={"source": "test"},
    )


@pytest.fixture
def f_prob(times, lats, lons, members):
    """Probabilistic forecast containing also a member dimension."""
    data = np.random.rand(len(members), len(times), len(lats), len(lons))
    return xr.DataArray(
        data,
        coords=[members, times, lats, lons],
        dims=["member", "time", "lat", "lon"],
        attrs={"source": "test"},
    )


@pytest.fixture
def f(f_prob):
    """Deterministic forecast matching observation o."""
    return f_prob.isel(member=0, drop=True)


# a vs. b in deterministic
@pytest.fixture
def a(o):
    return o


@pytest.fixture
def b(f):
    return f


# nan
@pytest.fixture
def a_rand_nan(a):
    """Masked"""
    return a.where(a < 0.5)


@pytest.fixture
def b_rand_nan(b):
    """Masked"""
    return b.where(b < 0.5)


@pytest.fixture
def a_fixed_nan(a):
    """Masked block"""
    a.data[:, 1:3, 1:3] = np.nan
    return a


@pytest.fixture
def b_fixed_nan(b):
    """Masked block"""
    b.data[:, 1:3, 1:3] = np.nan
    return b


# with zeros
@pytest.fixture
def a_with_zeros(a):
    """Zeros"""
    return a.where(a < 0.5, 0)


# dask
@pytest.fixture
def a_dask(a):
    """Chunked"""
    return a.chunk()


@pytest.fixture
def b_dask(b):
    """Chunked"""
    return b.chunk()


@pytest.fixture
def a_rand_nan_dask(a_rand_nan):
    """Chunked"""
    return a_rand_nan.chunk()


@pytest.fixture
def b_rand_nan_dask(b_rand_nan):
    """Chunked"""
    return b_rand_nan.chunk()


@pytest.fixture
def o_dask(o):
    return o.chunk()


@pytest.fixture
def f_prob_dask(f_prob):
    return f_prob.chunk()


# 1D time
@pytest.fixture
def a_1d(a):
    """Timeseries of a"""
    return a.isel(lon=0, lat=0, drop=True)


@pytest.fixture
def b_1d(b):
    """Timeseries of b"""
    return b.isel(lon=0, lat=0, drop=True)


@pytest.fixture
def a_1d_fixed_nan():
    time = xr.cftime_range("2000-01-01", "2000-01-03", freq="D")
    return xr.DataArray([3, np.nan, 5], dims=["time"], coords=[time])


@pytest.fixture
def b_1d_fixed_nan(a_1d_fixed_nan):
    b = a_1d_fixed_nan.copy()
    b.values = [7, 2, np.nan]
    return b


@pytest.fixture
def a_1d_with_zeros(a_with_zeros):
    """Timeseries of a with zeros"""
    return a_with_zeros.isel(lon=0, lat=0, drop=True)


# weights
@pytest.fixture
def weights_cos_lat(a):
    """Weighting array by cosine of the latitude."""
    return xr.ones_like(a) * np.abs(np.cos(a.lat))


@pytest.fixture
def weights_linear_time(a):
    """Weighting array by linear (1 -> 0) of the time."""
    weights = np.linspace(1, 0, num=len(a.time))
    return xr.ones_like(a) * xr.DataArray(weights, dims="time")


@pytest.fixture
def weights_linear_time_1d(weights_linear_time):
    """Timeseries of weights_linear_time"""
    return weights_linear_time.isel(lon=0, lat=0, drop=True)


@pytest.fixture
def weights_lonlat(a):
    weights = np.cos(np.deg2rad(a.lat))
    _, weights = xr.broadcast(a, weights)
    return weights.isel(time=0, drop=True)


@pytest.fixture
def weights_time():
    time = xr.cftime_range("2000-01-01", "2000-01-03", freq="D")
    return xr.DataArray([1, 2, 3], dims=["time"], coords=[time])


@pytest.fixture
def weights_cos_lat_dask(weights_cos_lat):
    """
    Weighting array by cosine of the latitude.
    """
    return weights_cos_lat.chunk()


@pytest.fixture
def category_edges():
    """Category bin edges between 0 and 1."""
    return np.linspace(0, 1, 6)


@pytest.fixture
def forecast_3d_int():
    """Random integer 3D forecast used for testing Contingency."""
    times = xr.cftime_range(start="2000", freq="D", periods=10)
    lats = np.arange(4)
    lons = np.arange(5)
    data = np.random.randint(0, 10, size=(len(times), len(lats), len(lons)))
    return xr.DataArray(data, coords=[times, lats, lons], dims=["time", "lat", "lon"])


@pytest.fixture
def observation_3d_int(forecast_3d_int):
    """Random integer 3D observation used for testing Contingency."""
    b = forecast_3d_int.copy()
    b.values = np.random.randint(0, 10, size=(b.shape[0], b.shape[1], b.shape[2]))
    return b


@pytest.fixture
def forecast_3d():
    """Random 3D forecast used for testing Contingency."""
    times = xr.cftime_range(start="2000", freq="D", periods=10)
    lats = np.arange(4)
    lons = np.arange(5)
    data = np.random.normal(0, 1, size=(len(times), len(lats), len(lons)))
    return xr.DataArray(data, coords=[times, lats, lons], dims=["time", "lat", "lon"])


@pytest.fixture
def observation_3d(forecast_3d):
    """Random 3D observation used for testing Contingency."""
    b = forecast_3d.copy()
    b.values = np.random.normal(0, 1, size=(b.shape[0], b.shape[1], b.shape[2]))
    return b


@pytest.fixture
def dichotomous_Contingency_1d():
    """Contingency of fixed, dichotomous 1d forecast and observations."""
    observations = xr.DataArray(
        np.array(2 * [0] + 2 * [1] + 1 * [0] + 2 * [1]), coords=[("x", np.arange(7))]
    )
    forecasts = xr.DataArray(
        np.array(2 * [0] + 2 * [0] + 1 * [1] + 2 * [1]), coords=[("x", np.arange(7))]
    )
    category_edges = np.array([-np.inf, 0.5, np.inf])
    return Contingency(
        observations, forecasts, category_edges, category_edges, dim=["x"]
    )


@pytest.fixture
def symmetric_edges():
    """Category bin edges between 0 and 1."""
    return np.linspace(-3, 3, 11)


@pytest.fixture
def forecast_1d_long():
    """Forecasts normally distributed around 0."""
    s = 100
    return xr.DataArray(np.random.normal(size=(s)), coords=[("time", np.arange(s))])


@pytest.fixture
def observation_1d_long():
    """Observations normally distributed around 0."""
    s = 100
    return xr.DataArray(np.random.normal(size=(s)), coords=[("time", np.arange(s))])



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import datetime
import subprocess
import sys

import sphinx_autosummary_accessors

import xskillscore

print("python exec:", sys.executable)
print("sys.path:", sys.path)

if "conda" in sys.executable:
    print("conda environment:")
    subprocess.run(["conda", "list"])
else:
    print("pip environment:")
    subprocess.run(["pip", "list"])

print("xskillscore: %s, %s" % (xskillscore.__version__, xskillscore.__file__))


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "IPython.sphinxext.ipython_directive",
    "IPython.sphinxext.ipython_console_highlighting",
    "nbsphinx",
    "sphinx_autosummary_accessors",
]

extlinks = {
    "issue": ("https://github.com/xarray-contrib/xskillscore/issues/%s", "GH#"),
    "pr": ("https://github.com/xarray-contrib/xskillscore/pull/%s", "GH#"),
}

autodoc_typehints = "none"

nbsphinx_timeout = 60
nbsphinx_execute = "always"

autosummary_generate = True

napoleon_use_param = True
napoleon_use_rtype = True

numpydoc_class_members_toctree = True
numpydoc_show_class_members = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates", sphinx_autosummary_accessors.templates_path]

# The suffix of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "xskillscore"
copyright = "2018-%s, xskillscore Developers" % datetime.datetime.now().year

# The full version, including alpha/beta/rc tags
version = xskillscore.__version__

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
today_fmt = "%Y-%m-%d"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["build", "**.ipynb_checkpoints", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Can add below once we have a logo.
# html_logo = 'images/esmtools-logo.png'
# html_theme_options = {'logo_only': True, 'style_nav_header_background': '#fcfcfc'}

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = today_fmt

# Output file base name for HTML help builder.
htmlhelp_basename = "xskillscoredoc"

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "cftime": ("https://unidata.github.io/cftime", None),
    "dask": ("https://docs.dask.org/en/latest", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "python": ("https://docs.python.org/3/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference", None),
    "sklearn": ("https://scikit-learn.org/stable", None),
    "xarray": ("https://xarray.pydata.org/en/stable", None),
}



import numpy as np
import pytest
import sklearn.metrics
from scipy.stats import linregress, pearsonr, spearmanr
from sklearn.metrics import (
    mean_absolute_error,
    mean_absolute_percentage_error,
    mean_squared_error,
    r2_score,
)

import xskillscore as xs
from xskillscore.core.deterministic import (
    linslope,
    mae,
    mape,
    me,
    mse,
    pearson_r,
    pearson_r_p_value,
    r2,
    rmse,
    smape,
    spearman_r,
    spearman_r_p_value,
)

xs_skl_metrics = [
    (r2, r2_score),
    (mse, mean_squared_error),
    (mae, mean_absolute_error),
    (mape, mean_absolute_percentage_error),
]

xs_skl_metrics_with_zeros = [
    (mape, mean_absolute_percentage_error),
]

xs_scipy_metrics = [
    (linslope, linregress, 0),
    (pearson_r, pearsonr, 0),
    (spearman_r, spearmanr, 0),
    (pearson_r_p_value, pearsonr, 1),
    (spearman_r_p_value, spearmanr, 1),
]


xs_np_metrics = [
    (me, lambda x, y: np.mean(x - y)),
    (smape, lambda x, y: np.mean(np.abs(x - y) / (np.abs(x) + np.abs(y)))),
]


@pytest.mark.parametrize("xs_skl_metrics", xs_skl_metrics)
def test_xs_same_as_skl(a_1d, b_1d, xs_skl_metrics):
    """Tests xskillscore metric is same as scikit-learn metric."""
    xs_metric, skl_metric = xs_skl_metrics
    actual = xs_metric(a_1d, b_1d, "time")
    expected = skl_metric(a_1d, b_1d)
    assert np.allclose(actual, expected)


def test_xs_same_as_skl_rmse(a_1d, b_1d):
    actual = rmse(a_1d, b_1d, "time")
    expected = mean_squared_error(a_1d, b_1d, squared=False)
    assert np.allclose(actual, expected)


def test_xs_same_as_skl_same_name(a_1d, b_1d):
    """Tests xskillscore metric is same as scikit-learn metric
    for metrics with same name."""
    xs_metric, skl_metric = (
        getattr(xs, "median_absolute_error"),
        getattr(sklearn.metrics, "median_absolute_error"),
    )
    actual = xs_metric(a_1d, b_1d, "time")
    expected = skl_metric(a_1d, b_1d)
    assert np.allclose(actual, expected)


@pytest.mark.parametrize("xs_skl_metrics", xs_skl_metrics_with_zeros)
def test_xs_same_as_skl_with_zeros(a_1d_with_zeros, b_1d, xs_skl_metrics):
    """Tests xskillscore metric is same as scikit-learn metric."""
    xs_metric, skl_metric = xs_skl_metrics
    actual = xs_metric(a_1d_with_zeros, b_1d, "time")
    expected = skl_metric(a_1d_with_zeros, b_1d)
    assert np.allclose(actual, expected)


@pytest.mark.parametrize("xs_scipy_metrics", xs_scipy_metrics)
def test_xs_same_as_scipy(a_1d, b_1d, xs_scipy_metrics):
    """Tests xskillscore metric is same as scipy metric."""
    xs_metric, scipy_metric, i = xs_scipy_metrics
    actual = xs_metric(a_1d, b_1d, "time")
    expected = scipy_metric(a_1d, b_1d)[i]
    assert np.allclose(actual, expected)


@pytest.mark.parametrize("xs_np_metrics", xs_np_metrics)
def test_xs_same_as_numpy(a_1d, b_1d, xs_np_metrics):
    """Tests xskillscore metric is same as metric using numpy."""
    xs_metric, np_metric = xs_np_metrics
    actual = xs_metric(a_1d, b_1d, "time")
    expected = np_metric(a_1d, b_1d)
    assert np.allclose(actual, expected)



# Write the benchmarking functions here.
# See "Writing benchmarks" in the asv docs for more information.


import os
import shutil

import bottleneck as bn
import numpy as np
import pandas as pd
import xarray as xr

from xskillscore import (
    mae as xs_mae,
    mse as xs_mse,
    pearson_r as xs_pearson_r,
    rmse as xs_rmse,
    spearman_r as xs_spearman_r,
)

from . import Generate, parameterized, requires_dask

# These metrics in xskillscore.xr.deterministic, entirely written in xarray functions,
# are identical to the well documented metrics in xskillscore.core.deterministic, which
# are based on numpy functions applied to xarray objects by xarray.apply_ufunc. As the
# xr metrics are only faster for small data, their use is not encouraged, as the
# numpy-based metrics are 20-40% faster on large data."""


def xr_mse(a, b, dim=None, skipna=True, weights=None):
    res = (a - b) ** 2
    if weights is not None:
        res = res.weighted(weights)
    res = res.mean(dim=dim, skipna=skipna)
    return res


def xr_mae(a, b, dim=None, skipna=True, weights=None):
    res = np.abs(a - b)
    if weights is not None:
        res = res.weighted(weights)
    res = res.mean(dim=dim, skipna=skipna)
    return res


def xr_me(a, b, dim=None, skipna=True, weights=None):
    res = a - b
    if weights is not None:
        res = res.weighted(weights)
    res = res.mean(dim=dim, skipna=skipna)
    return res


def xr_rmse(a, b, dim=None, skipna=True, weights=None):
    res = (a - b) ** 2
    if weights is not None:
        res = res.weighted(weights)
    res = res.mean(dim=dim, skipna=skipna)
    res = np.sqrt(res)
    return res


def xr_pearson_r(a, b, dim=None, **kwargs):
    return xr.corr(a, b, dim)


def _rankdata(o, dim):
    if isinstance(dim, str):
        dim = [dim]
    elif dim is None:
        dim = list(o.dims)
    if len(dim) == 1:
        return xr.apply_ufunc(
            bn.nanrankdata,
            o,
            input_core_dims=[[]],
            kwargs={"axis": o.get_axis_num(dim[0])},
            dask="allowed",
        )
    elif len(dim) > 1:
        # stack rank unstack
        return xr.apply_ufunc(
            bn.nanrankdata,
            o.stack(ndim=dim),
            input_core_dims=[[]],
            kwargs={"axis": -1},
            dask="allowed",
        ).unstack("ndim")


def xr_spearman_r(a, b, dim=None, **kwargs):
    return xr.corr(_rankdata(a, dim), _rankdata(b, dim), dim)


METRICS = [
    xs_mse,
    xr_mse,
    xs_rmse,
    xr_rmse,
    xs_mae,
    xr_mae,
    xs_pearson_r,
    xr_pearson_r,
    xs_spearman_r,
    xr_spearman_r,
]

DIMS = ["member", ["lon", "lat"]]

large_lon_lat = 2000
large_lon_lat_chunksize = large_lon_lat // 4
ntime = 24


class Compute_small(Generate):
    """
    A benchmark xskillscore.metric for small xr.DataArrays"""

    def setup(self, *args, **kwargs):
        self.make_ds(ntime, 1, 1)  # no grid

    @parameterized(["metric", "dim"], (METRICS, DIMS))
    def time_xskillscore_metric(self, metric, dim):
        """Take time for xskillscore.metric."""
        metric(self.ds["tos"], self.ds["sos"], dim=dim).compute()

    @parameterized(["metric", "dim"], (METRICS, DIMS))
    def peakmem_xskillscore_metric(self, metric, dim):
        """Take memory peak for xskillscore.metric."""
        metric(self.ds["tos"], self.ds["sos"], dim=dim).compute()


class Compute_large(Compute_small):
    """
    A benchmark xskillscore.metric for large xr.DataArrays"""

    def setup(self, *args, **kwargs):
        self.make_ds(ntime, large_lon_lat, large_lon_lat)


class Compute_large_dask(Compute_small):
    """
    A benchmark xskillscore.metric for large xr.DataArrays with dask."""

    def setup(self, *args, **kwargs):
        requires_dask()
        self.make_ds(
            ntime, large_lon_lat, large_lon_lat, chunks={"lon": large_lon_lat_chunksize}
        )



from setuptools import find_packages, setup

DISTNAME = "xskillscore"
LICENSE = "Apache"
AUTHOR = "Ray Bell"
AUTHOR_EMAIL = "rayjohnbell0@gmail.com"
DESCRIPTION = "xskillscore"
LONG_DESCRIPTION = """Metrics for verifying forecasts"""
URL = "https://github.com/xarray-contrib/xskillscore"
with open("requirements.txt") as f:
    INSTALL_REQUIRES = f.read().strip().split("\n")
TESTS_REQUIRE = ["pytest"]
PYTHON_REQUIRE = ">=3.7"


setup(
    name=DISTNAME,
    license=LICENSE,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    url=URL,
    packages=find_packages(),
    install_requires=INSTALL_REQUIRES,
    test_suite="xskillscore/tests",
    tests_require=TESTS_REQUIRE,
    python_requires=PYTHON_REQUIRE,
    use_scm_version={"version_scheme": "post-release", "local_scheme": "dirty-tag"},
    setup_requires=[
        "setuptools_scm",
        "setuptools>=30.3.0",
        "setuptools_scm_git_archive",
    ],
    zip_safe=False,
)
